#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
# pmx is Copyright (C) 2013-2020 by Daniel Seeliger, Vytautas Gapsys
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""Program to map ligand atoms for hybrid structure/topology generation.
"""

import sys
import argparse
from pmx import *
#from pmx.options import *
#from pmx.parser import *
from pmx.model import Model
from pmx.parser import read_and_format
from pmx.utils import get_ff_path
from pmx.ndx import *
from pmx.scripts.cli import check_unknown_cmd
from pmx.ligand_alchemy import *
import time
import warnings
from rdkit import Chem


# =============
# Input Options
# =============
def parse_options():

    parser = argparse.ArgumentParser(description='''
Provided two structures find atoms to be morphed.

''', formatter_class=argparse.RawTextHelpFormatter)

#    exclus = parser.add_mutually_exclusive_group()

    parser.add_argument('-i1',
                        metavar='lig1.pdb',
                        dest='i1',
                        type=str,
                        help='Input ligand structure 1. '
                        'Default is "lig1.pdb"',
                        default='lig1.pdb')
    parser.add_argument('-i2',
                        metavar='lig2.pdb',
                        dest='i2',
                        type=str,
                        help='Input ligand structure 2. '
                        'Default is "lig2.pdb"',
                        default='lig2.pdb')
    parser.add_argument('-o1',
                        metavar='pairs1.dat',
                        dest='o1',
                        type=str,
                        help='Output pairs: column1:mol1, column2:mol2. '
                        'Default is "pairs1.dat"',
                        default='pairs1.dat')
    parser.add_argument('-o2',
                        metavar='pairs2.dat',
                        dest='o2',
                        type=str,
                        help='Output pairs: column1:mol2, column2:mol1. '
                        'Default is "pairs2.dat"',
                        default='pairs2.dat')
    parser.add_argument('-opdb1',
                        metavar='out_pdb1.pdb',
                        dest='opdb1',
                        type=str,
                        help='Optional output: superimposed structure 1. ')
    parser.add_argument('-opdb2',
                        metavar='out_pdb2.pdb',
                        dest='opdb2',
                        type=str,
                        help='Optional output: superimposed structure 2. ')
    parser.add_argument('-opdbm1',
                        metavar='out_pdb_morphe1.pdb',
                        dest='opdbm1',
                        type=str,
                        help='Optional output: morphable atoms in structure 1. ')
    parser.add_argument('-opdbm2',
                        metavar='out_pdb_morphe2.pdb',
                        dest='opdbm2',
                        type=str,
                        help='Optional output: morphable atoms in structure 2. ')
    parser.add_argument('-score',
                        metavar='out_score.dat',
                        dest='score',
                        type=str,
                        help='Optional output: score of the morph. '
                        'Default is "out_score.dat"',
                        default='out_score.dat')
    parser.add_argument('-n1',
                        metavar='scaffold1.ndx',
                        dest='n1',
                        type=str,
                        help='Optional input: index of atoms to consider for mol1 ')
    parser.add_argument('-n2',
                        metavar='scaffold2.ndx',
                        dest='n2',
                        type=str,
                        help='Optional input: index of atoms to consider for mol2 ')
    parser.add_argument('-log',
                        metavar='mapping.log',
                        dest='log',
                        type=str,
                        help='Output: log file. '
                        'Default is "mapping.log"',
                        default='mapping.log')
    parser.add_argument('--no-alignment',
                        dest='alignment',
                        help='Should the alignment method be disabled (default enabled)',
                        action='store_false')
    parser.add_argument('--no-mcs',
                        dest='mcs',
                        help='Should the MCS method be disabled (default enabled)',
                        action='store_false')
    parser.add_argument('--no-H2H',
                        dest='H2H',
                        help='Should non-polar hydrogens be discarded from morphing into any other hydrogen (default True)',
                        action='store_false')
    parser.add_argument('--H2Hpolar',
                        dest='H2Hpolar',
                        help='Should polar hydrogens be morphed into polar hydrogens (default False)',
                        action='store_true')
    parser.add_argument('--H2Heavy',
                        dest='H2Heavy',
                        help='Should hydrogen be morphed into a heavy atom (default False)',
                        action='store_true')
    parser.add_argument('--RingsOnly',
                        dest='RingsOnly',
                        help='Should rings only be used in the MCS search and alignemnt (default False)',
                        action='store_true')
    parser.add_argument('--dMCS',
                        dest='dMCS',
                        help='Should the distance criterium be also applied in the MCS based search (default False)',
                        action='store_true')
    parser.add_argument('--swap',
                        dest='swap',
                        help='Try swapping the molecule order which would be a cross-check and require double execution time (default False)',
                        action='store_true')
    parser.add_argument('--no-chirality',
                        dest='chirality',
                        help='Perform chirality check for MCS mapping (default True)',
                        action='store_false')
    parser.add_argument('--d',
                        metavar='0.05',
                        dest='d',
                        type=float,
                        help='Distance (nm) between atoms to consider them morphable for alignment approach (default 0.05 nm).',
                        default=0.05)
    parser.add_argument('--timeout',
                        metavar='10',
                        dest='timeout',
                        type=int,
                        help='Maximum time (s) for an MCS search (default 10 s).',
                        default=10)

    parser.set_defaults(alignment=True,mcs=True,H2H=True,H2Hpolar=False,H2Heavy=False,
                        RingsOnly=False,dMCS=False,swap=False,chirality=True)
    args, unknown = parser.parse_known_args()
    check_unknown_cmd(unknown)

    return args

def entry_point():
    args = parse_options()
    if args.H2Heavy==True:
        args.H2H=True
    main(args)

def main(args):
    """Run the main script.

    Parameters
    ----------
    args : argparse.Namespace
        The command line arguments
    """

#####################################
    # start timing
    stime = time.time()

#####################################
    # deal with the flags
    bH2H = args.H2H
    bH2Hpolar = args.H2Hpolar
    bH2Heavy = args.H2Heavy
    bChiral = args.chirality
    bSwap = args.swap
    bRingsOnly = args.RingsOnly
    d = args.d
    timeout = args.timeout
    bdMCS = args.dMCS 

#####################################
    # log file
    logfile = open(args.log,'w')

#####################################
   # identify the methods to use 
    bAlignment = args.alignment
    bMCS = args.mcs
    if bMCS==False and bAlignment==False:
        doLog(logfile,"No method (alignment, mcs) was selected.",commandName='atomMapping')
        sys.exit(0)
    doLog(logfile,"Morphable atoms will be identified using the following methods:",commandName='atomMapping')
    doLog(logfile,"Alignment: {0}".format(bAlignment),commandName='atomMapping')
    doLog(logfile,"MCS: {0}".format(bMCS),commandName='atomMapping')

#####################################
    # read index
    read_from_idx1 = False
    read_from_idx2 = False
    if args.n1 is not None:
        read_from_idx1 = True
        idx1 = IndexFile(args.n1)
    if args.n2 is not None:
        read_from_idx2 = True
        idx2 = IndexFile(args.n2)

#####################################
    # reformat PDB to read two letter atoms properly (why does it have to be so painful?)
    if not os.path.exists( args.i1 ):
        doLog(logfile,'Input pdb1 not found. Exiting...',commandName='atomMapping')
        sys.exit(0)
    if not os.path.exists( args.i2 ):
        doLog(logfile,'Input pdb2 not found. Exiting...',commandName='atomMapping')
        sys.exit(0)
    pid = os.getpid()
    pdbName1,atomNameID1,sigmaHoleID1 = reformatPDB(args.i1,1,pid)
    pdbName2,atomNameID2,sigmaHoleID2 = reformatPDB(args.i2,2,pid)
    mol1 = Chem.MolFromPDBFile(pdbName1,removeHs=False,sanitize=False)
    mol2 = Chem.MolFromPDBFile(pdbName2,removeHs=False,sanitize=False)
    try:
        Chem.SanitizeMol(mol1)
    except:
        pass
    try:
        Chem.SanitizeMol(mol2)
    except:
        pass
    os.remove(pdbName1)
    os.remove(pdbName2)

#####################################
    ## prepare molecules for MCS ####
    molForMcs1 = cp.deepcopy(mol1)
    molForMcs2 = cp.deepcopy(mol2)
    try:
        rdmolops.AssignAtomChiralTagsFromStructure(mol1)
        rdmolops.AssignAtomChiralTagsFromStructure(mol2)
        molForMcs1 = cp.deepcopy(mol1)
        molForMcs2 = cp.deepcopy(mol2)
        rdmolops.AssignStereochemistry(mol1)
        rdmolops.AssignStereochemistry(mol2)
    except:
        doLog(logfile,"Chirality not assigned",commandName='atomMapping')

#####################################
    # deal with the -RingsOnly flag
    bYesRings = checkRingsOnlyFlag(mol1,mol2)
    if(bRingsOnly==True):
        if( bYesRings==True ):
            bRingsOnly=True
        else:
            doLog(logfile,"-RingsOnly flag is unset, because one (or both) molecule has no rings",commandName='atomMapping')
    
    n1 = []
    n2 = []

#########################
    # make molecule copies #
    molcp1 = cp.deepcopy(mol1)
    molcp2 = cp.deepcopy(mol2)

#########################
    # create an object of LigandsAtomMapping class
    doLog(logfile,"Initializing the main LigandsAtomMapping object",commandName='atomMapping')
    ligMap = LigandAtomMapping(mol1=mol1, mol2=mol2, molForMcs1=molForMcs1, molForMcs2=molForMcs2, bH2H=bH2H, bH2Hpolar=bH2Hpolar, bH2Heavy=bH2Heavy, bdMCS=bdMCS, bRingsOnly=bRingsOnly, d=d, bChiral=bChiral, sigmaHoleID1=sigmaHoleID1, sigmaHoleID2=sigmaHoleID2, timeout=timeout, logfile=logfile, bElements=False, bCarbonize=False, commandName='atomMapping' )



#########################
    # mcs
    bMCSfailed = False
    n1mcs = []
    n2mcs = []
    if(bMCS==True):
        doLog(logfile,"The topology matching approach will be used (MCS)",commandName='atomMapping')
        doLog(logfile,"fmcs module: Copyright (c) 2012 Andrew Dalke Scientific AB",commandName='atomMapping')
#########################
        # mcs: only rings
        if(bRingsOnly==True):
            doLog(logfile,"MCS: rings only...",commandName='atomMapping')
            ligMap.bElements = False
            ligMap.bCarbonize = False
            ligMap.bRingsOnly = True
            n1mcs,n2mcs = ligMap.mcs()
        else:
#########################
            # mcs: all atoms, match elements, modified molecules
            n1A = []
            n2A = []
            ligMap.bElements = True
            ligMap.bCarbonize = True
            ligMap.bRingsOnly = False
            doLog(logfile,"MCS: trying to run an MCS using all atoms: modified molecules, match elements...",commandName='atomMapping')
            n1A,n2A = ligMap.mcs()
            doLog(logfile,"MCS: all atoms, modified molecules, size of mapping: {0}".format(len(n1A)),commandName='atomMapping')

#########################
            # mcs: all atoms, match elements, unmodified molecules
            ligMap.bElements = True
            ligMap.bCarbonize = False
            ligMap.bRingsOnly = False
            doLog(logfile,"MCS: trying to run an MCS using all atoms: unmodified molecules, match elements...",commandName='atomMapping')
            n1B,n2B = ligMap.mcs()
            doLog(logfile,"MCS: all atoms, unmodified molecules, size of mapping: {0}".format(len(n1B)),commandName='atomMapping')

#########################
            # mcs: select the best mapping
            n1mcs,n2mcs = ligMap._compare_mappings_by_size( mol1,mol2, n1A, n2A, n1B, n2B )

#########################
            # mcs: rings only
            if( bYesRings==True ):
                doLog(logfile,"MCS: trying to run an MCS using rings only...",commandName='atomMapping')
                ligMap.bElements = False
                ligMap.bCarbonize = False
                ligMap.bRingsOnly = True
                n1C,n2C = ligMap.mcs()
                n1mcs,n2mcs = ligMap._compare_mappings_by_size( mol1,mol2, n1mcs, n2mcs, n1C, n2C ) 

        doLog(logfile,"MCS: using variant with the mapping size of {0}".format(len(n1mcs)),commandName='atomMapping')

        if len(n1mcs)==0:
             bMCSfailed = True

#########################
    # alignment
    n1align = []
    n2align = []
    if( (bAlignment==True) or (bMCSfailed==True) ):
        if bMCSfailed==True:
            doLog(logfile,"The MCS approach did not find any mapping",commandName='atomMapping')
        doLog(logfile,"The alignment approach will be used",commandName='atomMapping')
        doLog(logfile,"Tosco, P., Balle, T. & Shiri, F. Open3DALIGN: an open-source software aimed at unsupervised ligand alignment. J Comput Aided Mol Des 25:777-83 (2011)",commandName='atomMapping')
        doLog(logfile,"Alignment is based on atom logP contributions: S. A. Wildman and G. M. Crippen JCICS _39_ 868-873 (1999)",commandName='atomMapping')

#########################
    # alignment: rings only
        if(bRingsOnly==True):
            ligMap.bRingsOnly = True
            n1align,n2align,pyO3A = ligMap.alignment()
            doLog(logfile,"Alignment: size of mapping: {0}".format(len(n1align)),commandName='atomMapping')
        else:
#########################
            # alignment: all atoms
            doLog(logfile,"Alignment: trying to align all atoms...",commandName='atomMapping')
            ligMap.bRingsOnly = False
            n1align,n2align,pyO3A = ligMap.alignment()
            doLog(logfile,"Alignment: size of mapping: {0}".format(len(n1align)),commandName='atomMapping')
#########################
            # alignment: rings only
            if( bYesRings==True ):
                doLog(logfile,"Alignment: trying to align rings only...",commandName='atomMapping')
                ligMap.bRingsOnly = True
                mol1 = cp.deepcopy(molcp1)
                mol2 = cp.deepcopy(molcp2)
                n1Balign,n2Balign,pyO3A = ligMap.alignment()
                doLog(logfile,"Size of mapping: {0}".format(len(n1Balign)),commandName='atomMapping')

#########################
                # alignment: select the best mapping
                if(len(n1align)<=len(n1Balign)):
                    doLog(logfile,"Alignment: using ring only alignment result.",commandName='atomMapping')
                    n1align = n1Balign
                    n2align = n2Balign
                else:
                    doLog(logfile,"Alignment: using all atom alignment result.",commandName='atomMapping')

#########################
    # select the best mapping: mcs vs alignment
    if len(n1align)>=len(n1mcs):
        doLog(logfile,"The final result is based on the Alignment.",commandName='atomMapping')
        n1 = n1align
        n2 = n2align
    else:
        doLog(logfile,"The final result is based on the MCS.\n",commandName='atomMapping')
        n1 = n1mcs
        n2 = n2mcs

#########################
    # swap TODO


       
#########################
    # check
    if( len(n1) != len(n2) ):
        doLog(logfile,"ERROR: something went wrong.",commandName='atomMapping')
        doLog(logfile,"ERROR: Number of the morphable atoms in the ligands does not match.",commandName='atomMapping')
        sys.exit(1)

#########################
    # calculate score
    score = ligMap._calc_score(n1,n2,mol1,mol2)

#########################
    # report results
    doLog(logfile,"Final results:",commandName='atomMapping')
    if( bH2H==True or bH2Heavy==True ):
        doLog(logfile,"Atoms considered in mol1: {0}".format(mol1.GetNumAtoms()),commandName='atomMapping')
        doLog(logfile,"Atoms considered in mol2: {0}".format(mol2.GetNumAtoms()),commandName='atomMapping')
    else:
        doLog(logfile,"Atoms considered in mol1: {0}".format(mol1.GetNumHeavyAtoms()),commandName='atomMapping')
        doLog(logfile,"Atoms considered in mol2: {0}".format(mol2.GetNumHeavyAtoms()),commandName='atomMapping')
    doLog(logfile,"Morphable atoms in both molecules: {0} {1}".format(len(n1),len(n2)),commandName='atomMapping')
    doLog(logfile,"Dissimilarity (distance) score: %.4f\n" % score,commandName='atomMapping')
    if args.score!=None:
        fp = open(args.score,'w')
        fp.write("Score: %.4f\n" % score)
        fp.close()

#########################
    # restore atom names
    restoreAtomNames(mol1,atomNameID1)
    restoreAtomNames(mol2,atomNameID2)

#########################
    # output pdb
    if args.opdb1 != None:
        Chem.MolToPDBFile(molcp1,args.opdb1)
    if args.opdb2 != None:
        try:
            Chem.rdMolAlign.AlignMol(mol2,molcp1,atomMap=list(zip(n2,n1)))
        except:
            doLog(logfile,"PDB output: cannot superimpose -opdb2 structure. Maybe no morphable atoms have been found",commandName='atomMapping')
        Chem.MolToPDBFile(mol2,args.opdb2)
    if args.opdbm1!=None:
        mol = submol_by_index(molcp1,n1)
        Chem.MolToPDBFile(mol,args.opdbm1)
    if args.opdbm2!=None:
        mol = submol_by_index(mol2,n2)
        Chem.MolToPDBFile(mol,args.opdbm2)

#########################
    # output pairs
    ligMap._write_pairs(n1,n2,args.o1)
    ligMap._write_pairs(n2,n1,args.o2)


if __name__ == '__main__':
    entry_point()




import numpy as np
import sys
from scipy.optimize import fmin
import scipy.stats
from .utils import data2gauss

__all__ = ['Jarz', 'JarzGauss', 'Crooks', 'BAR']

# Constants
kb = 0.00831447215   # kJ/(K*mol)


class Jarz:
    '''Jarzynski estimator. [1]_

    Both the forward and reverse estimates, as well as the mean of the two
    are calculated.

    The standard error is calculated via bootstrap when ``nboots``>0, and by
    separating the work values into groups when ``nblocks``>1.

    Parameters
    ----------
    wf : array_like
        array of forward work values.
    wr : array_like
        array of reverse work values.
    T : float, optional
        temperature in Kelvin. Default is 298.15 K.
    nboots : int, optional
        how many bootstrap samples to draw for estimating the standard error.
        Default is zero (do not estimate the error).
    nblocks : int, optional
        how many blocks to divide the input work values into for the estimation
        of the standard error. Default is one (do not estimate the error).
    statesProvided : string
        both directions 'AB', forward only 'A', reverse only 'B'

    Examples
    --------
    >>> estimate = Jarz(wf, wr, T=300, nboots=1000, nblocks=10)
    >>> dg_forward = estimate.dg_for
    >>> dg_reverse = estimate.dg_rev
    >>> dg_mean = estimate.dg_mean
    >>> dg_forward_err1 = estimate.err_boot_for
    >>> dg_forward_err2 = estimate.err_blocks_for

    Attributes
    ----------
    dg_for : float
        the forward free energy estimate.
    dg_rev : float
        the reverse free energy estimate.
    dg_mean : float
        the mean of the forward and reverse free energy estimates.
    err_boot_for : float
        standard error of the forward free energy estimate calculated
        via bootstrap.
    err_boot_rev : float
        standard error of the reverse free energy estimate calculated
        via bootstrap.
    err_blocks_for : float
        standard error of the forward free energy estimate calculated by
        separating the input work values into groups/blocks.
    err_blocks_rev : float
        standard error of the reverse free energy estimate calculated by
        separating the input work values into groups/blocks.
    '''

    def __init__(self, wf=[], wr=[], T=298.15, nboots=0, nblocks=1, statesProvided='AB'):
        if 'A' in statesProvided:
            self.wf = np.array(wf)
        if 'B' in statesProvided:
            self.wr = np.array(wr)
        self.T = float(T)
        self.nboots = nboots
        self.nblocks = nblocks

        # Calculate all Jarz properties available
        if 'A' in statesProvided:
            self.dg_for = self.calc_dg(w=self.wf, T=self.T, bReverse=False)
        if 'B' in statesProvided:
            self.dg_rev = self.calc_dg(w=self.wr, T=self.T, bReverse=True)
        if 'AB' in statesProvided:
            self.dg_mean = (self.dg_for + self.dg_rev) * 0.5

        if nboots > 0:
            if 'A' in statesProvided:
                self.err_boot_for = self.calc_err_boot(w=self.wf, T=self.T,
                                                   nboots=self.nboots,
                                                   bReverse=False)
            if 'B' in statesProvided:
                self.err_boot_rev = self.calc_err_boot(w=self.wr, T=self.T,
                                                   nboots=self.nboots,
                                                   bReverse=True)

        if nblocks > 1:
            if 'A' in statesProvided:
                self.err_blocks_for = self.calc_err_blocks(w=self.wf,
                                                       T=self.T,
                                                       nblocks=self.nblocks,
                                                       bReverse=False)
            if 'B' in statesProvided:
                self.err_blocks_rev = self.calc_err_blocks(w=self.wr,
                                                       T=self.T,
                                                       nblocks=self.nblocks,
                                                       bReverse=True)

    @staticmethod
    def calc_dg(w, T, bReverse=False):
        '''Calculates the free energy difference using Jarzynski's equality.
        [1]_

        Parameters
        ----------
        w : array_like
            array of work values.
        T : float
            temperature in Kelvin.
        bReverse : bool, optional
            whether the work values provided are for the reverse transition.
            Default if False. If they are for the reverse transition, set it to
            True.

        Returns
        -------
        dg : float
            estimate of the free energy difference.
        '''

        if bReverse is False:
            c = 1.0
        elif bReverse is True:
            c = -1.0

        beta = 1./(kb*T)
        n = float(len(w))
        mexp = 0.0
        m = 0.0
        m2 = 0.0

        w = np.array(w)
        median = np.median(w)
        for i in w:
            mexp = mexp + np.exp( -beta*c*i - (-beta*c*median) )
#            mexp = mexp + np.exp(-beta*c*i)

        # Jarzynski estimator
        dg = c*median -kb*T*np.log(mexp) + kb*T*np.log(n)

        return c * dg

    @staticmethod
    def calc_err_boot(w, T, nboots, bReverse=False):
        '''Calculates the standard error via bootstrap. The work values are
        resampled randomly with replacement multiple (nboots) times,
        and the Jarzinski free energy recalculated for each bootstrap samples.
        The standard error of the estimate is returned as the standard d
        eviation of the bootstrapped free energies.

        Parameters
        ----------
        w : array_like
            work values.
        T : float
            temperature.
        bReverse : bool, optional
            whether the work values provided are for the reverse transition.
            Default if False. If they are for the reverse transition, set it to
            True.
        nboots: int
            number of bootstrap samples to use for the error estimate.

        Returns
        ----------
        sderr : float
            the standard error of the estimate.
        '''
        dg_boots = []
        n = len(w)
        for k in range(nboots):
            sys.stdout.write('\r  Bootstrap (Std Err): iteration %s/%s'
                             % (k+1, nboots))
            sys.stdout.flush()

            boot = np.random.choice(w, size=n, replace=True)
            dg_boot = Jarz.calc_dg(w=boot, T=T, bReverse=bReverse)
            dg_boots.append(dg_boot)
        sys.stdout.write('\n')
        err = np.std(dg_boots)
        return err

    @staticmethod
    def calc_err_blocks(w, T, nblocks, bReverse=False):
        '''Calculates the standard error based on a number of blocks the
        work values are divided into. It is useful when you run independent
        equilibrium simulations, so that you can then use their respective
        work values to compute the standard error based on the repeats.

        Parameters
        ----------
        w : array_like
            array of work values.
        T : float
            temperature.
        bReverse : bool, optional
            whether the work values provided are for the reverse transition.
            Default if False. If they are for the reverse transition, set it to
            True.
        nblocks: int
            number of blocks to divide the data into. This can be for
            instance the number of independent equilibrium simulations
            you ran.

        Returns
        -------
        sderr : float
            the standard error of the estimate.
        '''

        dg_blocks = []
        # loosely split the arrays
        w_split = np.array_split(w, nblocks)

        # calculate all dg
        for w_block in w_split:
            dg_block = Jarz.calc_dg(w=w_block, T=T, bReverse=bReverse)
            dg_blocks.append(dg_block)

        # get std err
        err_blocks = scipy.stats.sem(dg_blocks, ddof=1)

        return err_blocks


class JarzGauss:
    '''Jarzynski estimator using a Gaussian approximation. [6]_

    Both the forward and reverse estimates.

    The standard error is calculated using the analytical expression derived
    by Hummer (2001). [6]_ When ``nboots``>0, the error is also calculated
    by bootstrap. When  ``nblocks``>1, it is calculated by
    separating the work values into blocks.

    Parameters
    ----------
    wf : array_like
        array of forward work values.
    wr : array_like
        array of reverse work values.
    T : float, optional
        temperature in Kelvin. Default is 298.15 K.
    nboots : int, optional
        how many bootstrap samples to draw for estimating the standard error.
        Default is zero (do not estimate the error).
    nblocks : int, optional
        how many blocks to divide the input work values into for the estimation
        of the standard error. Default is one (do not estimate the error).
    statesProvided: str, optional
        two directions or one

    Examples
    --------
    >>> estimate = JarzGauss(wf, wr, T=300, nboots=1000, nblocks=10)
    >>> dg_forward = estimate.dg_for
    >>> dg_reverse = estimate.dg_rev
    >>> dg_forward_err1 = estimate.err_for
    >>> dg_forward_err2 = estimate.err_boot_for
    >>> dg_forward_err3 = estimate.err_blocks_for

    Attributes
    ----------
    dg_for : float
        the forward free energy estimate.
    dg_rev : float
        the reverse free energy estimate.
    err_for : float
        analytical estimate of the standard error of the forward free energy
        estimate.
    err_rev : float
        analytical estimate of the standard error of the reverse free energy
        estimate.
    err_boot_for : float
        standard error of the forward free energy estimate calculated
        via bootstrap.
    err_boot_rev : float
        standard error of the reverse free energy estimate calculated
        via bootstrap.
    err_blocks_for : float
        standard error of the forward free energy estimate calculated by
        separating the input work values into groups/blocks.
    err_blocks_rev : float
        standard error of the reverse free energy estimate calculated by
        separating the input work values into groups/blocks.
    '''

    def __init__(self, wf=[], wr=[], T=298.15, nboots=0, nblocks=1, statesProvided='AB'):
        if 'A' in statesProvided:
            self.wf = np.array(wf)
        if 'B' in statesProvided:
            self.wr = np.array(wr)
        self.T = float(T)
        self.nboots = nboots
        self.nblocks = nblocks

        # Calculate all Jarz properties available
        if 'A' in statesProvided:
            self.dg_for = self.calc_dg(w=self.wf, T=self.T, bReverse=False)
            self.err_for = self.calc_err(w=self.wf, T=self.T, bReverse=False)
        if 'B' in statesProvided:
            self.dg_rev = self.calc_dg(w=self.wr, T=self.T, bReverse=True)
            self.err_rev = self.calc_err(w=self.wr, T=self.T, bReverse=True)

        if nboots > 0:
            if 'A' in statesProvided:
                self.err_boot_for = self.calc_err_boot(w=self.wf, T=self.T,
                                                   nboots=self.nboots,
                                                   bReverse=False)
            if 'B' in statesProvided:
                self.err_boot_rev = self.calc_err_boot(w=self.wr, T=self.T,
                                                   nboots=self.nboots,
                                                   bReverse=True)

        if nblocks > 1:
            if 'A' in statesProvided:
                self.err_blocks_for = self.calc_err_blocks(w=self.wf,
                                                       T=self.T,
                                                       nblocks=self.nblocks,
                                                       bReverse=False)
            if 'B' in statesProvided:
                self.err_blocks_rev = self.calc_err_blocks(w=self.wr,
                                                       T=self.T,
                                                       nblocks=self.nblocks,
                                                       bReverse=True)

    @staticmethod
    def calc_dg(w, T, bReverse=False):
        '''Calculates the free energy difference using the Jarzynski estimator
        with a Gaussian approximation. [6]_

        Parameters
        ----------
        w : array_like
            array of work values.
        T : float
            temperature in Kelvin.
        bReverse : bool, optional
            whether the work values provided are for the reverse transition.
            Default if False. If they are for the reverse transition, set it to
            True.

        Returns
        -------
        dg : float
            estimate of the free energy difference.
        '''
        beta = 1./(kb*T)
        if bReverse is False:
            c = 1.0
        elif bReverse is True:
            c = -1.0

        dg = np.mean(c*w) - (beta * np.var(c*w, ddof=1)) * 0.5
        return c * dg

    @staticmethod
    def calc_err(w, T, bReverse=False):
        '''Calculates the standard error via an analytic expression.
        The expression is derived by Hummer, 2001, JChemPhys. [6]_

        Parameters
        ----------
        w : array_like
            work values.
        T : float
            temperature.
        bReverse : bool, optional
            whether the work values provided are for the reverse transition.
            Default if False. If they are for the reverse transition, set it to
            True.

        Returns
        -------
        err : float
            standard deviation of the estimator.
        '''

        beta = 1./(kb*T)
        w_var = np.var(w, ddof=1)
        n = float(len(w))
        dg_var = w_var/n + np.power(beta*w_var, 2) / (2.0 * (n-1.0))
        dg_stderr = np.sqrt(dg_var)

        return dg_stderr

    @staticmethod
    def calc_err_boot(w, T, nboots, bReverse=False):
        '''Calculates the standard error via bootstrap. The work values are
        resampled randomly with replacement multiple (nboots) times,
        and the Gaussian approximation for Jarzinski free energy
        is recalculated for each bootstrap samples.
        The standard error of the estimate is returned as the standard d
        eviation of the bootstrapped free energies.

        Parameters
        ----------
        w : array_like
            work values.
        T : float
            temperature.
        nboots: int
            number of bootstrap samples to use for the error estimate.
        bReverse : bool, optional
            whether the work values provided are for the reverse transition.
            Default if False. If they are for the reverse transition, set it to
            True.

        Returns
        -------
        err : float
            standard error of the mean.
        '''
        dg_boots = []
        n = len(w)
        for k in range(nboots):
            sys.stdout.write('\r  Bootstrap (Std Err): iteration %s/%s'
                             % (k+1, nboots))
            sys.stdout.flush()

            boot = np.random.choice(w, size=n, replace=True)
            dg_boot = JarzGauss.calc_dg(boot, T, bReverse)
            dg_boots.append(dg_boot)
        sys.stdout.write('\n')
        err = np.std(dg_boots)
        return err

    @staticmethod
    def calc_err_blocks(w, T, nblocks, bReverse=False):
        '''Calculates the standard error based on a number of blocks the
        work values are divided into. It is useful when you run independent
        equilibrium simulations, so that you can then use their respective
        work values to compute the standard error based on the repeats.

        Parameters
        ----------
        w : array_like
            array of work values.
        T : float
            temperature.
        nblocks: int
            number of blocks to divide the data into. This can be for
            instance the number of independent equilibrium simulations
            you ran.
        bReverse : bool, optional
            whether the work values provided are for the reverse transition.
            Default if False. If they are for the reverse transition, set it to
            True.

        Returns
        -------
        sderr : float
            the standard error of the estimate.
        '''

        dg_blocks = []
        # loosely split the arrays
        w_split = np.array_split(w, nblocks)

        # calculate all dg
        for w_block in w_split:
            dg_block = JarzGauss.calc_dg(w_block, T, bReverse)
            dg_blocks.append(dg_block)

        # get std err
        err_blocks = scipy.stats.sem(dg_blocks, ddof=1)

        return err_blocks


class Crooks:
    '''Crooks Gaussian Intersection (CGI) estimator. [2]_

    The forward and reverse work
    values are fitted to Gaussian functions and their intersection is taken
    as the free energy estimate. In some cases, when the two Gaussians are very
    close to each other, the intersection cannot be taken and the average of
    the two Gaussian means is taken as the free energy estimate insted.

    The standard error is calculated via bootstrap when ``nboots``>0, and by
    separating the work values into groups when ``nblocks``>1. Bootstrap errors
    are calculating with two different bootstrap procedures: a parametric one
    in which samples are drawn from Gaussian distributions fitted to the
    original work distributions, and a nonparametric one in which the work
    values themself are drawn at random with replacement.

    Parameters
    ----------
    wf : array_like
        array of forward work values.
    wr : array_like
        array of reverse work values.
    nboots : int, optional
        how many bootstrap samples to draw for estimating the standard error.
        Default is zero (do not estimate the error).
    nblocks : int, optional
        how many blocks to divide the input work values into for the estimation
        of the standard error. Default is one (do not estimate the error).

    Examples
    --------
    >>> estimate = Crooks(wf, wr, nboots=1000, nblocks=10)
    >>> dg = estimate.dg
    >>> dg_err1 = estimate.err_boot1
    >>> dg_err2 = estimate.err_boot2
    >>> dg_err3 = estimate.err_blocks

    Attributes
    ----------
    dg : float
        the free energy estimate.
    err_boot1 : float
        standard error of the free energy estimate calculated via parametric
        bootstrap.
    err_boot2 : float
        standard error of the free energy estimate calculated via
        nonparametric bootstrap.
    err_blocks : float
        standard error of the free energy estimate calculated by separating
        the input work values into groups/blocks.
    inters_bool : bool
        whether the interection could be taken. If False, the free energy
        estimate is the average of the two Gaussian means.
    Af : float
        height of the forward Gaussian.
    mf : float
        mean of the forward Gaussian.
    devf : float
        standard deviation of the forward Gaussian.
    Ar : float
        height of the reverse Gaussian.
    mr : float
        mean of the reverse Gaussian.
    devr : float
        standard deviation of the reverse Gaussian.

    '''

    def __init__(self, wf, wr, nboots=0, nblocks=1):

        # inputs
        self.wf = np.array(wf)
        self.wr = np.array(wr)
        self.nboots = nboots
        self.nblocks = nblocks
        # params of the gaussians
        self.mf, self.devf, self.Af = data2gauss(wf)
        self.mr, self.devr, self.Ar = data2gauss(wr)

        # Calculate Crooks properties
        self.dg, self.inters_bool = self.calc_dg(wf=self.wf, wr=self.wr)

        if nboots > 0:
            self.err_boot1 = self.calc_err_boot1(m1=self.mf, s1=self.devf,
                                                 n1=len(wf), m2=self.mr,
                                                 s2=self.devr, n2=len(wr),
                                                 nboots=nboots)
            self.err_boot2 = self.calc_err_boot2(wf=self.wf, wr=self.wr,
                                                 nboots=nboots)

        if nblocks > 1:
            self.err_blocks = self.calc_err_blocks(self.wf, self.wr, nblocks)

    @staticmethod
    def calc_dg(wf, wr):
        '''Calculates the free energy difference using the Crooks Gaussian
        Intersection method. It finds the intersection of two Gaussian
        functions. If the intersection cannot be computed, the average of
        the two Gaussian locations is returned.

        Parameters
        ----------
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.

        Returns
        -------
        float
            location of the intersection.
        bool
            whether the intersection could be calculated. If the intersection
            was calculated as expected a True value is returned.
            If the Gaussians are too close to each other, the intersection
            cannot be calculated and a False value is returned; in this case,
            the first float value retured is the average of the Gaussian means.
        '''

        m1, s1, A1 = data2gauss(wf)
        m2, s2, A2 = data2gauss(wr)

        p1 = m1/s1**2-m2/s2**2
        p2 = np.sqrt(1/(s1**2*s2**2)*(m1-m2)**2+2*(1/s1**2-1/s2**2)*np.log(s2/s1))
        p3 = 1/s1**2-1/s2**2
        x1 = (p1+p2)/p3
        x2 = (p1-p2)/p3

        # determine which solution to take
        if x1 > m1 and x1 < m2 or x1 > m2 and x1 < m1:
            dg = x1
            return dg, True
        elif x2 > m1 and x2 < m2 or x2 > m2 and x2 < m1:
            dg = x2
            return dg, True
        else:
            # we do not take the intersection but the average of the means
            dg = (m1 + m2) * 0.5
            return dg, False

    # Possible change of behaviour compared to the original script:
    # here it is not determined in advanced whether to take the intersection
    # or the mean, but for each bootstrap sample if the intersecion cannot
    # be taken, then the mean is used automatically.
    @staticmethod
    def calc_err_boot1(m1, s1, n1, m2, s2, n2, nboots):
        '''Calculates the standard error of the Crooks Gaussian Intersection
        via parametric bootstrap. Given the parameters of the forward and
        reverse Gaussian distributions, multiple (nboots) bootstrap samples
        are built by random sampling from these two Gaussian distributions.
        The CGI free energy is then calculated for each bootstrap sample
        (forward and reverse Gaussians). The standard error of the estimate
        is returned as the standard deviation of the bootstrapped free
        energies.

        Parameters
        ----------
        m1 : float
            mean of the forward Gaussian.
        s1 : float
            standard deviation of the forward Gaussian.
        n1 : int
            number of work values to which the first Gaussian was fit.
        m2 : float
            mean of the reverse Gaussian.
        s2 : float
            standard deviation of the reverse Gaussian.
        n2 : int
            number of work values to which the second Gaussian was fit.
        nboots: int
            number of bootstrap samples to use for the error estimate.
            Parametric bootstrap is used where work values are resampled from
            two Gaussians.

        Returns
        ----------
        sderr : float
            the standard error of the estimate.
        '''

        dg_boots = []
        for k in range(nboots):
            bootA = np.random.normal(loc=m1, scale=s1, size=n1)
            bootB = np.random.normal(loc=m2, scale=s2, size=n2)

            dg_boot, _ = Crooks.calc_dg(bootA, bootB)
            dg_boots.append(dg_boot)
        err = np.std(dg_boots)
        return err

    @staticmethod
    def calc_err_boot2(wf, wr, nboots):
        '''Calculates the standard error of the Crooks Gaussian Intersection
        via non-parametric bootstrap. The work values are resampled randomly
        with replacement multiple (nboots) times, and the CGI free energy
        recalculated for each bootstrap samples. The standard error of
        the estimate is returned as the standard deviation of the bootstrapped
        free energies.

        Parameters
        ----------
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        nboots: int
            number of bootstrap samples to use for the error estimate.

        Returns
        ----------
        sderr : float
            the standard error of the estimate.
        '''

        nf = len(wf)
        nr = len(wr)

        dg_boots = []
        for k in range(nboots):
            sys.stdout.write('\r  Bootstrap (Std Err): iteration %s/%s'
                             % (k+1, nboots))
            sys.stdout.flush()

            bootA = np.random.choice(wf, size=nf, replace=True)
            bootB = np.random.choice(wr, size=nr, replace=True)

            dg_boot, _ = Crooks.calc_dg(bootA, bootB)
            dg_boots.append(dg_boot)

        sys.stdout.write('\n')
        err = np.std(dg_boots)
        return err

    @staticmethod
    def calc_err_blocks(wf, wr, nblocks):
        '''Calculates the standard error based on a number of blocks the
        work values are divided into. It is useful when you run independent
        equilibrium simulations, so that you can then use their respective
        work values to compute the standard error based on the repeats.

        Parameters
        ----------
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        nblocks: int
            number of blocks to divide the data into. This can be for
            instance the number of independent equilibrium simulations
            you ran.

        Returns
        ----------
        sderr : float
            the standard error of the estimate.
        '''

        dg_blocks = []
        # loosely split the arrays
        wf_split = np.array_split(wf, nblocks)
        wr_split = np.array_split(wr, nblocks)

        # calculate all dg
        for wf_block, wr_block in zip(wf_split, wr_split):
            dg_block, _ = Crooks.calc_dg(wf_block, wr_block)
            dg_blocks.append(dg_block)

        # get std err
        err_blocks = scipy.stats.sem(dg_blocks, ddof=1)

        return err_blocks


class BAR:
    '''Bennett acceptance ratio (BAR) estimator. [3]_

    The standard error is calculated via bootstrap when ``nboots``>0, and by
    separating the work values into groups when ``nblocks``>1.

    Parameters
    ----------
    wf : array_like
        array of forward work values.
    wr : array_like
        array of reverse work values.
    T : float, optional
        temperature in Kelvin. Default is 298.15 K.
    nboots : int, optional
        how many bootstrap samples to draw for estimating the standard error.
        Default is zero (do not estimate the error).
    nblocks : int, optional
        how many blocks to divide the input work values into for the estimation
        of the standard error. Default is one (do not estimate the error).

    Examples
    --------
    >>> estimate = BAR(wf, wr, T=300, nboots=1000, nblocks=10)
    >>> dg = estimate.dg
    >>> dg_err1 = estimate.err
    >>> dg_err2 = estimate.err_boot
    >>> dg_err3 = estimate.err_blocks
    >>> dg_convergece = estimate.conv

    Attributes
    ----------
    dg : float
        the free energy estimate.
    err : float
        analytical estimate of the standard error.
    err_boot : float
        standard error of the free energy estimate calculated via bootstrap.
    err_blocks : float
        standard error of the free energy estimate calculated by separating
        the input work values into groups/blocks.
    conv : float
        convergence of the free energy estimate. [4]_
    conv_err_boot : float
        standard error of the convergence estimate calculated via bootstrap.

    '''

    def __init__(self, wf, wr, T, nboots=0, nblocks=1):
        self.wf = np.array(wf)
        self.wr = np.array(wr)
        self.T = float(T)
        self.nboots = nboots
        self.nblocks = nblocks

        self.nf = len(wf)
        self.nr = len(wr)
        self.beta = 1./(kb*self.T)
        self.M = kb * self.T * np.log(float(self.nf) / float(self.nr))

        # Calculate all BAR properties available
        self.dg = self.calc_dg(self.wf, self.wr, self.T)
        self.err = self.calc_err(self.dg, self.wf, self.wr, self.T)
        self.conv = self.calc_conv(self.dg, self.wf, self.wr, self.T)
        if nboots > 0:
            self.err_boot = self.calc_err_boot(self.wf, self.wr, nboots,
                                               self.T)
            self.conv_err_boot = self.calc_conv_err_boot(self.dg, self.wf,
                                                         self.wr, nboots,
                                                         self.T)
        if nblocks > 1:
            self.err_blocks = self.calc_err_blocks(self.wf, self.wr, nblocks,
                                                   self.T)
            self.conv_err_blocks = self.calc_conv_err_blocks(self.dg, self.wf,
                                                         self.wr, nblocks,
                                                         self.T)

    @staticmethod
    def calc_dg(wf, wr, T):
        '''Estimates and returns the free energy difference.

        Parameters
        ----------
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        T : float
            temperature

        Returns
        ----------
        dg : float
            the BAR free energy estimate.
        '''

        nf = float(len(wf))
        nr = float(len(wr))
        beta = 1./(kb*T)
        M = kb * T * np.log(nf/nr)

        def func(x, wf, wr):
            sf = 0
            for v in wf:
                sf += 1./(1+np.exp(beta*(M+v-x)))

            sr = 0
            for v in wr:
                sr += 1./(1+np.exp(-beta*(M+v-x)))

            r = sf-sr
            return r**2

        avA = np.average(wf)
        avB = np.average(wr)
        x0 = (avA+avB)/2.
        dg = fmin(func, x0=x0, args=(wf, wr), disp=0)

        return float(dg)

    @staticmethod
    def calc_err(dg, wf, wr, T):
        '''Calculates the analytical error estimate.

        Parameters
        ----------
        dg : float
            the BAR free energy estimate
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        T : float
            temperature

        Returns
        ----------
        sderr : float
            the standard error of the estimate.
        '''

        nf = float(len(wf))
        nr = float(len(wr))
        beta = 1./(kb*T)
        M = kb * T * np.log(nf/nr)

        err = 0
        for v in wf:
            err += 1./(2+2*np.cosh(beta * (M+v-dg)))
        for v in wr:
            err += 1./(2+2*np.cosh(beta * (M+v-dg)))
        N = nf + nr
        err /= float(N)
        tot = 1/(beta**2*N)*(1./err-(N/nf + N/nr))

        err = float(np.sqrt(tot))
        return err

    @staticmethod
    def calc_err_boot(wf, wr, nboots, T):
        '''Calculates the error by bootstrapping.

        Parameters
        ----------
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        T : float
            temperature
        nboots: int
            number of bootstrap samples.

        Returns
        ----------
        sderr : float
            the standard error of the estimate.
        '''

        nf = len(wf)
        nr = len(wr)
        dg_boots = []
        for k in range(nboots):
            sys.stdout.write('\r  Bootstrap (Std Err): iteration %s/%s'
                             % (k+1, nboots))
            sys.stdout.flush()

            bootA = np.random.choice(wf, size=nf, replace=True)
            bootB = np.random.choice(wr, size=nr, replace=True)
            dg_boot = BAR.calc_dg(bootA, bootB, T)
            dg_boots.append(dg_boot)

        sys.stdout.write('\n')
        err_boot = np.std(dg_boots)

        return err_boot

    @staticmethod
    def calc_err_blocks(wf, wr, nblocks, T):
        '''Calculates the standard error based on a number of blocks the
        work values are divided into. It is useful when you run independent
        equilibrium simulations, so that you can then use their respective
        work values to compute the standard error based on the repeats.

        Parameters
        ----------
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        T : float
            temperature
        nblocks: int
            number of blocks to divide the data into. This can be for
            instance the number of independent equilibrium simulations
            you ran.

        Returns
        ----------
        sderr : float
            the standard error of the estimate.
        '''

        dg_blocks = []
        # loosely split the arrays
        wf_split = np.array_split(wf, nblocks)
        wr_split = np.array_split(wr, nblocks)

        # calculate all dg
        for wf_block, wr_block in zip(wf_split, wr_split):
            dg_block = BAR.calc_dg(wf_block, wr_block, T)
            dg_blocks.append(dg_block)

        # get std err
        err_blocks = scipy.stats.sem(dg_blocks, ddof=1)

        return err_blocks

    @staticmethod
    def calc_conv(dg, wf, wr, T):
        '''Evaluates BAR convergence as described by Hahn & Then. [4]_

        Returns a value between -1 and 1: the closer this
        value to zero the better the BAR convergence.

        Parameters
        ----------
        dg : float
            the BAR free energy estimate
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        T : float
            temperature

        Returns
        ----------
        conv : float
            convergence estimate.
        '''

        wf = np.array(wf)
        wr = np.array(wr)

        beta = 1./(kb*T)
        nf = len(wf)
        nr = len(wr)
        N = float(nf + nr)

        ratio_alpha = float(nf)/N
        ratio_beta = float(nr)/N
        bf = 1.0/(ratio_beta + ratio_alpha * np.exp(beta*(wf-dg)))
        tf = 1.0/(ratio_alpha + ratio_beta * np.exp(beta*(-wr+dg)))
        Ua = (np.mean(tf) + np.mean(bf))/2.0
        Ua2 = (ratio_alpha * np.mean(np.power(tf, 2)) +
               ratio_beta * np.mean(np.power(bf, 2)))
        conv = (Ua-Ua2)/Ua
        return conv

    @staticmethod
    def calc_conv_err_boot(dg, wf, wr, nboots, T):
        '''Calculates the error in the convergence measure by bootstrapping.

        Parameters
        ----------
        dg : float
            the BAR free energy estimate
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        T : float
            temperature
        nboots: int
            number of bootstrap samples.

        Returns
        ----------
        sderr : float
            the standard error of the convergence measure.
        '''

        nf = len(wf)
        nr = len(wr)
        conv_boots = []
        for k in range(nboots):
            sys.stdout.write('\r  Bootstrap (Conv): '
                             'iteration %s/%s' % (k+1, nboots))
            sys.stdout.flush()

            bootA = np.random.choice(wf, size=nf, replace=True)
            bootB = np.random.choice(wr, size=nr, replace=True)
            conv_boot = BAR.calc_conv(dg, bootA, bootB, T)
            conv_boots.append(conv_boot)

        sys.stdout.write('\n')
        err = np.std(conv_boots)
        return err

    @staticmethod
    def calc_conv_err_blocks(dg, wf, wr, nblocks, T):
        '''Calculates the standard error of the convergence measure
        based on a number of blocks the work values are divided into. 
        It is useful when you run independent
        equilibrium simulations, so that you can then use their respective
        work values to compute the standard error based on the repeats.

        Parameters
        ----------
        dg : float
            the BAR free energy estimate
        wf : array_like
            array of forward work values.
        wr : array_like
            array of reverse work values.
        T : float
            temperature
        nblocks: int
            number of blocks to divide the data into. This can be for
            instance the number of independent equilibrium simulations
            you ran.

        Returns
        ----------
        sderr : float
            the standard error of the convergence measure.
        '''

        conv_blocks = []
        # loosely split the arrays
        wf_split = np.array_split(wf, nblocks)
        wr_split = np.array_split(wr, nblocks)

        # calculate all dg
        for wf_block, wr_block in zip(wf_split, wr_split):
            conv_block = BAR.calc_conv(dg, wf_block, wr_block, T)
            conv_blocks.append(conv_block)

        # get std err
        err_blocks = scipy.stats.sem(conv_blocks, ddof=1)

        return err_blocks




#!/usr/bin/env python

from pmx.parser import read_and_format
from pmx.estimators import Jarz, JarzGauss, Crooks, BAR
from pmx.analysis import read_dgdl_files, plot_work_dist, ks_norm_test
from pmx.utils import natural_sort
from pmx import __version__
import sys
import os
import time
import numpy as np
import pickle
import argparse
import warnings
from pmx.scripts.cli import check_unknown_cmd

# Constants
kb = 0.00831447215   # kJ/(K*mol)


# ==============================================================================
#                               FUNCTIONS
# ==============================================================================
def _dump_integ_file(outfn, f_lst, w_lst):
    with open(outfn, 'w') as f:
        for fn, w in zip(f_lst, w_lst):
            f.write('{dhdl} {work}\n'.format(dhdl=fn, work=w))


def _data_from_file(fn):
    data = read_and_format(fn, 'sf')
    return map(lambda a: a[1], data)


def _tee(fp, s, quiet=False):
    print(s, file=fp)
    if quiet is False:
        print(s)


def _time_stats(seconds):
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return h, m, s


# ==============================================================================
#                      COMMAND LINE OPTIONS AND MAIN
# ==============================================================================
def parse_options():

    parser = argparse.ArgumentParser(description='Calculates free energies '
            'from fast growth thermodynamic integration simulations. '
            'Available methods for free energy estimation: '
            'Crooks Gaussian Intersection (CGI); '
            'Benett Acceptance Ratio (BAR); '
            'Jarzinski equality (JARZ); '
            'Jarzinski with a Gaussian approximatiuon (JARZ_Gauss).')

    exclu1 = parser.add_mutually_exclusive_group()
    # the following exclusion groups are for fA/iA and fB/iB:
    # one can either specify -fA or -iA, but one of them is required.
    exclu2 = parser.add_mutually_exclusive_group()#required=True)
    exclu3 = parser.add_mutually_exclusive_group()#required=True)

    exclu2.add_argument('-fA',
                        metavar='dhdl',
                        dest='filesAB',
                        type=str,
                        help='dhdl.xvg files for the A->B simulations. Use '
                        'wildcard to select multiple xvg files: e.g. "-fa '
                        './forward_results/dhdl*.xvg"',
                        nargs='+')
    exclu3.add_argument('-fB',
                        metavar='dhdl',
                        dest='filesBA',
                        type=str,
                        help='dhdl.xvg files for the B->A simulations Use '
                        'wildcard to select multiple xvg files: e.g. "-fb '
                        './backward_results/dhdl*.xvg"',
                        nargs='+')
    parser.add_argument('-m',
                        metavar='method',
                        type=str.lower,
                        dest='methods',
                        help='Choose one or more estimators to use from the '
                        'available ones: CGI, BAR, JARZ. Default is all.',
                        default=['cgi', 'bar', 'jarz'],
                        nargs='+')
    parser.add_argument('-t',
                        metavar='temperature',
                        dest='temperature',
                        type=float,
                        help='Temperature in Kelvin. Default is 298.15.',
                        default=298.15)
    parser.add_argument('-o',
                        metavar='result file',
                        dest='outfn',
                        type=str,
                        help='Filename of output result file. Default is '
                        '"results.txt."',
                        default='results.txt')
    parser.add_argument('-b',
                        metavar='nboots',
                        dest='nboots',
                        type=int,
                        help='Number of bootstrap samples to use for the '
                        'bootstrap estimate of the standard errors. Default '
                        'is 0 (no bootstrap).',
                        default=100)
    parser.add_argument('-n',
                        metavar='nblocks',
                        dest='nblocks',
                        type=int,
                        help='Number of blocks to divide the data into for '
                        'an estimate of the standard error. You can use this '
                        'when multiple independent equilibrium simulations'
                        'have been run so to estimate the error from the '
                        'repeats. Default is 1 (i.e. no repeats). It assumes '
                        'the dhdl files for each repeat are read in order and '
                        'are contiguous, e.g. dhdl_0 to dhdl_9 is the first '
                        'repeat, dhdl_10 to dhdl_19 is the second one, etc.',
                        default=1)
    parser.add_argument('-w',
                        metavar='plot',
                        dest='wplot',
                        type=str,
                        help='Name of image file showing the distribution of '
                        'work values. Default is "wplot.png". If you want to '
                        'avoid saving this plot, pass "none" to this flag. '
                        'If you choose to calculate the free energy with '
                        'multiple estimators, the dG values shown on the plot '
                        'will be chosen following the hierarchy '
                        'BAR > CGI > JARZ.',
                        default='wplot.png')
    parser.add_argument('--nbins',
                        metavar='',
                        dest='nbins',
                        type=int,
                        help='Number of histograms bins for the plot. '
                        'Default is 20.',
                        default=20)
    parser.add_argument('--dpi',
                        metavar='',
                        dest='dpi',
                        type=int,
                        help='Resolution of the plot. Default is 300.',
                        default=300)
    parser.add_argument('--reverseB',
                        dest='reverseB',
                        help='Whether to reverse the work values for the '
                        'backward (B->A) transformation. This is useful '
                        'when in Gromacs both forward and reverse simulations '
                        'were run from lambda zero to one. '
                        'Default is False.',
                        default=False,
                        action='store_true')
    parser.add_argument('--integ_only',
                        dest='integ_only',
                        help='Whether to do integration only; the integrated '
                        'values are computed and saved, and the program '
                        'terminated. Default is False.',
                        default=False,
                        action='store_true')
    exclu2.add_argument('-iA',
                        metavar='work input',
                        dest='iA',
                        type=str,
                        help='Two-column dat file containing the list of input'
                        ' files and their respective integrated work values '
                        'for the forward (A->B) tranformation.')
    exclu3.add_argument('-iB',
                        metavar='work input',
                        dest='iB',
                        type=str,
                        help='Two-column dat file containing the list of input'
                        ' files and their respective integrated work values '
                        'for the reverse (B->A) tranformation.')
    parser.add_argument('-oA',
                        metavar='work output',
                        dest='oA',
                        type=str,
                        help='File where to save the list of input dhdl'
                        ' files and their respective integrated work values '
                        'for the forward (A->B) tranformation. Default is '
                        '"integA.dat"',
                        default='integA.dat')
    parser.add_argument('-oB',
                        metavar='work output',
                        dest='oB',
                        type=str,
                        help='File where to save the list of input dhdl'
                        ' files and their respective integrated work values '
                        'for the reverse (B->A) tranformation. Default is '
                        '"integB.dat"',
                        default='integB.dat')
    # The following are mutually exclusive options
    exclu1.add_argument('--skip',
                        metavar='',
                        dest='skip',
                        type=int,
                        help='Skip files, i.e. pick every nth work value. '
                        'Default is 1 (all); with 2, every other work value '
                        'is discarded, etc.',
                        default=1)
    exclu1.add_argument('--slice',
                        metavar='',
                        dest='slice',
                        type=int,
                        help='Subset of trajectories to analyze. '
                        'Provide list slice, e.g. "10 50" will '
                        'result in selecting dhdl_files[10:50]. '
                        'Default is all.',
                        default=None,
                        nargs=2)
    exclu1.add_argument('--rand',
                        metavar='',
                        dest='rand',
                        type=int,
                        help='Take a random subset of trajectories. '
                        'Default is None (do not take random subset)',
                        default=None)
    exclu1.add_argument('--index',
                        metavar='',
                        dest='index',
                        type=int,
                        help='Zero-based index of files to analyze (e.g.'
                        ' 0 10 20 50 60). It keeps '
                        'the dhdl.xvg files according to their position in '
                        'the list, sorted according to the filenames. '
                        'Default is None (i.e. all dhdl are used).',
                        default=None,
                        nargs='+')
    parser.add_argument('--prec',
                        metavar='',
                        dest='precision',
                        type=int,
                        help='The decimal precision of the screen/file output.'
                        ' Default is 2.',
                        default=2)
    parser.add_argument('--units',
                        metavar='',
                        dest='units',
                        type=str.lower,
                        help='The units of the output. Choose from "kJ", '
                        '"kcal", "kT". Default is "kJ."',
                        default='kJ',
                        choices=['kj', 'kcal', 'kt'])
    parser.add_argument('--pickle',
                        dest='pickle',
                        help='Whether to save the free energy results from '
                        'the estimators in pickled files. Default is False.',
                        default=False,
                        action='store_true')
    parser.add_argument('--no_ks',
                        dest='do_ks_test',
                        help='Whether to do a Kolmogorov-Smirnov test '
                        'to check whether the Gaussian assumption for CGI '
                        'holds. Default is True; this flag turns it to False.',
                        default=True,
                        action='store_false')
    parser.add_argument('--quiet',
                        dest='quiet',
                        help='Minimal screen output.',
                        default=False,
                        action='store_true')
    parser.add_argument('--sigmoid',
                        metavar='',
                        dest='sigmoid',
                        type=float,
                        help='Define a sigmoidal lambda path.',
                        default=0.0)

    args, unknown = parser.parse_known_args()
    check_unknown_cmd(unknown)

    return args


# ==============================================================================
#                               FUNCTIONS
# ==============================================================================
def main(args):
    """Run the main script.

    Parameters
    ----------
    args : argparse.Namespace
        The command line arguments
    """

    # start timing
    stime = time.time()

    # input arguments
    out = open(args.outfn, 'w')
    T = args.temperature
    sigmoid = args.sigmoid
    skip = args.skip
    prec = args.precision
    methods = args.methods
    reverseB = args.reverseB
    integ_only = args.integ_only
    nboots = args.nboots
    nblocks = args.nblocks
    do_ks_test = args.do_ks_test
    quiet = args.quiet

    #---------- find out whether two or only one direction is provided --------
    filesAB = []
    filesBA = []
    statesProvided = 'AB'
    if (args.iA is None) and (args.iB is None): # no integrated values provided
        if (args.filesAB is None) and (args.filesBA is None): 
            exit('Need to provide dhdl.xvg files or integrated work values')
        elif args.filesAB is None:
                statesProvided = 'B'
                _tee(out, 'Only one directional Jarzynski estimator will be used')
                filesBA = natural_sort(args.filesBA)
        elif args.filesBA is None: 
                statesProvided = 'A'
                _tee(out, 'Only one directional Jarzynski estimator will be used')
                filesAB = natural_sort(args.filesAB)
        else:
            filesAB = natural_sort(args.filesAB)
            filesBA = natural_sort(args.filesBA)
    elif (args.iA is None) or (args.iB is None):
        if args.iA is None:
            statesProvided = 'B'
        else:
            statesProvided = 'A'
        _tee(out, 'Only one directional Jarzynski estimator will be used')

    # -------------------
    # Select output units
    # -------------------
    units = args.units
    if units.lower() == 'kj':
        # kJ is the input from GMX
        unit_fact = 1.
        units = 'kJ/mol'
    elif units == 'kcal':
        unit_fact = 1./4.184
        units = 'kcal/mol'
    elif units.lower() == 'kt':
        unit_fact = 1./(kb*T)
        units = 'kT'
    else:
        exit('No unit type \'%s\' available' % units)

    print("# analyze_dhdl.py, pmx version = %s" % __version__, file=out)
    print("# pwd = %s" % os.getcwd(), file=out)
    print("# %s (%s)" % (time.asctime(), os.environ.get('USER')), file=out)
    print("# command = %s" % ' '.join(sys.argv), file=out)
    _tee(out, "\n", quiet=quiet)

    # ==========
    # Parse Data
    # ==========

    # If list of dhdl.xvg files are provided, parse dhdl
    # --------------------------------------------------
    res_ab = []
    res_ba = []
    if args.iA is None and args.iB is None:
        # If random selection is chosen, do this before reading files and
        # calculating the work values.
        if args.rand is not None:
            if 'A' in statesProvided:
                filesAB = np.random.choice(filesAB, size=args.rand, replace=False)
            if 'B' in statesProvided:
                filesBA = np.random.choice(filesBA, size=args.rand, replace=False)
            _tee(out, 'Selected random subset of %d trajectories.' % args.rand,
                 quiet=quiet)

        # If slice values provided, select the files needed. Again before
        # reading files so speed up the process
        if args.slice is not None:
            first = args.slice[0]
            last = args.slice[1]
            _tee(out, ' First trajectory read: %i' % first, quiet=quiet)
            _tee(out, ' Last trajectory read: %i' % last, quiet=quiet)
            _tee(out, '', quiet=quiet)
            if 'A' in statesProvided:
                filesAB = filesAB[first:last]
            if 'B' in statesProvided:
                filesBA = filesBA[first:last]

        # If index values provided, select the files needed
        if args.index is not None:
            # Avoid index out of range error if "wrong" indices are provided
            if 'A' in statesProvided:
                filesAB = [filesAB[i] for i in args.index if i < len(filesAB)]
            if 'B' in statesProvided:
                filesBA = [filesBA[i] for i in args.index if i < len(filesBA)]
            # ...but warn if this happens
            if any(i > (len(filesAB) - 1) for i in args.index):
                warnings.warn('\nindex out of range for some of your chosen '
                              '\nindices for the forward work values. This means you are'
                              '\ntrying to select input files that are not present.')
            if any(i > (len(filesBA) - 1) for i in args.index):
                warnings.warn('\nindex out of range for some of your chosen'
                              '\nindices for the reverse work values. This means you are'
                              '\ntrying to select input files that are not present.')

        # when skipping start count from end: in this way the last frame is
        # always included, and what can change is the first one
        if 'A' in statesProvided:
            filesAB = list(reversed(filesAB[::-skip]))
        if 'B' in statesProvided:
            filesBA = list(reversed(filesBA[::-skip]))

        # --------------------
        # Now read in the data
        # --------------------
        if quiet is True:
            print('Processing the input data...')
        elif quiet is False:
            print(' ========================================================')
            print('                   PROCESSING THE DATA')
            print(' ========================================================')
            print('  Forward Data')
        if 'A' in statesProvided:
            res_ab = read_dgdl_files(filesAB, lambda0=0,
                                 invert_values=False, verbose=not quiet, sigmoid=sigmoid)
            _dump_integ_file(args.oA, filesAB, res_ab)

        if quiet is False:
            print('  Reverse Data')
        if 'B' in statesProvided:
            res_ba = read_dgdl_files(filesBA, lambda0=1,
                                 invert_values=reverseB, verbose=not quiet, sigmoid=sigmoid)
            _dump_integ_file(args.oB, filesBA, res_ba)

    # If integrated work values are given instead, read those
    # -------------------------------------------------------
    elif args.iA is not None and args.iB is not None:
        if 'A' in statesProvided:
            print('  Reading integrated values (A->B) from', args.iA)
            res_ab.extend(_data_from_file(args.iA))
        if 'B' in statesProvided:
            print('  Reading integrated values (B->A) from', args.iB)
            res_ba.extend(_data_from_file(args.iB))
        # If slice values provided, select the files needed.
        if args.slice is not None:
            first = args.slice[0]
            last = args.slice[1]
            _tee(out, '  Keeping work values from %i to %i.' % (first, last))
            _tee(out, '')
            if 'A' in statesProvided:
                res_ab = res_ab[first:last]
            if 'B' in statesProvided:
                res_ba = res_ba[first:last]

    # If a mix of dhdl.xvg files and integrated work values are given
    # ---------------------------------------------------------------
    # TODO: we are currently not handling this - do we want to add it?
    elif (args.iA is not None and args.filesBA is not None) or (args.iB is not None and args.filesAB is not None):
        raise ValueError('you need to provide either -fA and -fB, or -iA and '
                         '-iB. Mixing dhdl.xvg files and integrated work '
                         'values is not currently supported.')
    # The following code should never get executed, as the case where the input
    # is not provided should have already by handled by argparse within
    # parse_options
    else:
        raise ValueError('you need to provide dhdl.xvg files or integrated '
                         'work values for both forward (-fA or -iA) ',
                         'and reverse (-fB or -iB) transisions.')

    # If asked to only do the integration of dhdl.xvg, exit
    if integ_only is True:
        if quiet is False:
            print('\n    Integration done. Skipping analysis.')
            print('\n    ......done........\n')
        exit()

    # ==============
    # Begin Analysis
    # ==============
    _tee(out, ' ========================================================', quiet=quiet)
    _tee(out, '                       ANALYSIS', quiet=quiet)
    _tee(out, ' ========================================================', quiet=quiet)
    if 'A' in statesProvided:
        _tee(out, '  Number of forward (0->1) trajectories: %d' % len(res_ab), quiet=quiet)
    if 'B' in statesProvided:
        _tee(out, '  Number of reverse (1->0) trajectories: %d' % len(res_ba), quiet=quiet)
    _tee(out, '  Temperature : %.2f K' % T, quiet=quiet)

    # ============================
    # Crooks Gaussian Intersection
    # ============================
    if 'cgi' in methods and 'AB' in statesProvided:
        _tee(out, '\n --------------------------------------------------------', quiet=quiet)
        _tee(out, '             Crooks Gaussian Intersection     ', quiet=quiet)
        _tee(out, ' --------------------------------------------------------', quiet=quiet)

        if quiet is True:
            print('Running CGI analysis...')
        elif quiet is False:
            print('  Calculating Intersection...')

        cgi = Crooks(wf=res_ab, wr=res_ba, nboots=nboots, nblocks=nblocks)
        if args.pickle is True:
            pickle.dump(cgi, open("cgi_results.pkl", "wb"))

        _tee(out, '  CGI: Forward Gauss mean = {m:8.{p}f} {u} '
                  'std = {s:8.{p}f} {u}'.format(m=cgi.mf*unit_fact,
                                                s=cgi.devf*unit_fact,
                                                p=prec, u=units),
             quiet=quiet)
        _tee(out, '  CGI: Reverse Gauss mean = {m:8.{p}f} {u} '
                  'std = {s:8.{p}f} {u}'.format(m=cgi.mr*unit_fact,
                                                s=cgi.devr*unit_fact,
                                                p=prec, u=units),
             quiet=quiet)

        if cgi.inters_bool is False:
            _tee(out, '\n  Gaussians too close for intersection calculation', quiet=quiet)
            _tee(out, '   --> Taking difference of mean values', quiet=quiet)

        _tee(out, '  CGI: dG = {dg:8.{p}f} {u}'.format(dg=cgi.dg*unit_fact,
                                                       p=prec, u=units), quiet=quiet)

        if nboots > 0:
            _tee(out, '  CGI: Std Err (bootstrap:parametric) = {e:8.{p}f} {u}'.format(e=cgi.err_boot1*unit_fact,
                                                                                      p=prec, u=units), quiet=quiet)
            _tee(out, '  CGI: Std Err (bootstrap) = {e:8.{p}f} {u}'.format(e=cgi.err_boot2*unit_fact,
                                                                           p=prec, u=units), quiet=quiet)

        if nblocks > 1:
            _tee(out, '  CGI: Std Err (blocks) = {e:8.{p}f} {u}'.format(e=cgi.err_blocks*unit_fact,
                                                                        p=prec, u=units), quiet=quiet)

    # --------------
    # Normality test
    # --------------
    if do_ks_test and 'AB' in statesProvided:
        if quiet is False:
            print('\n  Running KS-test...')
        q0, lam00, check0, bOk0 = ks_norm_test(res_ab)
        q1, lam01, check1, bOk1 = ks_norm_test(res_ba)

        _tee(out, '    Forward: gaussian quality = %3.2f' % q0, quiet=quiet)
        if bOk0:
            _tee(out, '             ---> KS-Test Ok', quiet=quiet)
        else:
            _tee(out, '             ---> KS-Test Failed. sqrt(N)*Dmax = %4.2f,'
                      ' lambda0 = %4.2f' % (q0, check0), quiet=quiet)
        _tee(out, '    Reverse: gaussian quality = %3.2f' % q1, quiet=quiet)
        if bOk1:
            _tee(out, '             ---> KS-Test Ok', quiet=quiet)
        else:
            _tee(out, '             ---> KS-Test Failed. sqrt(N)*Dmax = %4.2f,'
                      ' lambda0 = %4.2f' % (q1, check1), quiet=quiet)

    # ========================
    # Bennett Acceptance Ratio
    # ========================
    if 'bar' in methods and 'AB' in statesProvided:
        _tee(out, '\n --------------------------------------------------------', quiet=quiet)
        _tee(out, '             Bennett Acceptance Ratio     ', quiet=quiet)
        _tee(out, ' --------------------------------------------------------', quiet=quiet)

        if quiet is True:
            print('Running BAR analysis...')
        elif quiet is False:
            print('  Running Nelder-Mead Simplex algorithm... ')

        bar = BAR(res_ab, res_ba, T=T, nboots=nboots, nblocks=nblocks)
        if args.pickle:
            pickle.dump(bar, open("bar_results.pkl", "wb"))

        _tee(out, '  BAR: dG = {dg:8.{p}f} {u}'.format(dg=bar.dg*unit_fact, p=prec, u=units), quiet=quiet)
        _tee(out, '  BAR: Std Err (analytical) = {e:8.{p}f} {u}'.format(e=bar.err*unit_fact, p=prec, u=units), quiet=quiet)

        if nboots > 0:
            _tee(out, '  BAR: Std Err (bootstrap)  = {e:8.{p}f} {u}'.format(e=bar.err_boot*unit_fact, p=prec, u=units), quiet=quiet)
        if nblocks > 1:
            _tee(out, '  BAR: Std Err (blocks)  = {e:8.{p}f} {u}'.format(e=bar.err_blocks*unit_fact, p=prec, u=units), quiet=quiet)

        _tee(out, '  BAR: Conv = %8.2f' % bar.conv, quiet=quiet)

        if nboots > 0:
            _tee(out, '  BAR: Conv Std Err (bootstrap) = %8.2f' % bar.conv_err_boot, quiet=quiet)
        if nblocks > 1:
            _tee(out, '  BAR: Conv Std Err (blocks) = %8.2f' % bar.conv_err_blocks, quiet=quiet)

    # =========
    # Jarzynski
    # =========
    if 'jarz' in methods:
        _tee(out, '\n --------------------------------------------------------', quiet=quiet)
        _tee(out, '             Jarzynski estimator     ', quiet=quiet)
        _tee(out, ' --------------------------------------------------------', quiet=quiet)

        if quiet is True:
            print('Running Jarz analysis...')
        jarz = Jarz(wf=res_ab, wr=res_ba, T=T, nboots=nboots, nblocks=nblocks, statesProvided=statesProvided)
        if args.pickle:
            pickle.dump(jarz, open("jarz_results.pkl", "wb"))

        if 'A' in statesProvided:
            _tee(out, '  JARZ: dG Forward = {dg:8.{p}f} {u}'.format(dg=jarz.dg_for*unit_fact,
                                                                p=prec, u=units), quiet=quiet)
        if 'B' in statesProvided:
            _tee(out, '  JARZ: dG Reverse = {dg:8.{p}f} {u}'.format(dg=jarz.dg_rev*unit_fact,
                                                                p=prec, u=units), quiet=quiet)
        if 'AB' in statesProvided:
            _tee(out, '  JARZ: dG Mean    = {dg:8.{p}f} {u}'.format(dg=jarz.dg_mean*unit_fact,
                                                                p=prec, u=units), quiet=quiet)
        if nboots > 0:
            if 'A' in statesProvided:
                _tee(out, '  JARZ: Std Err Forward (bootstrap) = {e:8.{p}f} {u}'.format(e=jarz.err_boot_for*unit_fact, p=prec, u=units), quiet=quiet)
            if 'B' in statesProvided:
                _tee(out, '  JARZ: Std Err Reverse (bootstrap) = {e:8.{p}f} {u}'.format(e=jarz.err_boot_rev*unit_fact, p=prec, u=units), quiet=quiet)

        if nblocks > 1:
            if 'A' in statesProvided:
                _tee(out, '  JARZ: Std Err Forward (blocks) = {e:8.{p}f} {u}'.format(e=jarz.err_blocks_for*unit_fact, p=prec, u=units), quiet=quiet)
            if 'B' in statesProvided:
                _tee(out, '  JARZ: Std Err Reverse (blocks) = {e:8.{p}f} {u}'.format(e=jarz.err_blocks_rev*unit_fact, p=prec, u=units), quiet=quiet)

        # -------------------------------------
        # Jarzynski with Gaussian approximation
        # -------------------------------------
        if quiet is True:
            print('Running Jarzynski Gaussian approximation analysis...')
        jarzGauss = JarzGauss(wf=res_ab, wr=res_ba, T=T, nboots=nboots, nblocks=nblocks, statesProvided=statesProvided)
        if args.pickle:
            pickle.dump(jarzGauss, open("jarz_gauss_results.pkl", "wb"))

        if 'A' in statesProvided:
            _tee(out, '  JARZ_Gauss: dG Forward = {dg:8.{p}f} {u}'.format(dg=jarzGauss.dg_for*unit_fact, p=prec, u=units), quiet=quiet)
        if 'B' in statesProvided:
            _tee(out, '  JARZ_Gauss: dG Reverse = {dg:8.{p}f} {u}'.format(dg=jarzGauss.dg_rev*unit_fact,
p=prec, u=units), quiet=quiet)
        if 'AB' in statesProvided:
            _tee(out, '  JARZ_Gauss: dG Mean    = {dg:8.{p}f} {u}'.format(dg=(jarzGauss.dg_for+jarzGauss.dg_rev)/2.0*unit_fact, p=prec, u=units), quiet=quiet)
        if 'A' in statesProvided:
            _tee(out, '  JARZ_Gauss: Std Err (analytical) Forward = {dg:8.{p}f} {u}'.format(dg=jarzGauss.err_for*unit_fact, p=prec, u=units), quiet=quiet)
        if 'B' in statesProvided:
            _tee(out, '  JARZ_Gauss: Std Err (analytical) Reverse = {dg:8.{p}f} {u}'.format(dg=jarzGauss.err_rev*unit_fact, p=prec, u=units), quiet=quiet)
        if nboots > 0:
            if 'A' in statesProvided:
                _tee(out, '  JARZ_Gauss: Std Err Forward (bootstrap) = {e:8.{p}f} {u}'.format(e=jarzGauss.err_boot_for*unit_fact, p=prec, u=units), quiet=quiet)
            if 'B' in statesProvided:
                _tee(out, '  JARZ_Gauss: Std Err Reverse (bootstrap) = {e:8.{p}f} {u}'.format(e=jarzGauss.err_boot_rev*unit_fact, p=prec, u=units), quiet=quiet)

        if nblocks > 1:
            if 'A' in statesProvided:
                _tee(out, '  JARZ_Gauss: Std Err Forward (blocks) = {e:8.{p}f} {u}'.format(e=jarzGauss.err_blocks_for*unit_fact, p=prec, u=units), quiet=quiet)
            if 'B' in statesProvided:
                _tee(out, '  JARZ_Gauss: Std Err Reverse (blocks) = {e:8.{p}f} {u}'.format(e=jarzGauss.err_blocks_rev*unit_fact, p=prec, u=units), quiet=quiet)

    _tee(out, ' ========================================================', quiet=quiet)
    
    #close the outputfile, forcing a flush to disk, so that the unit test can read it properly
    out.close()

    # -----------------------
    # plot work distributions
    # -----------------------
    if args.wplot.lower() != 'none':
        if quiet is False:
            print('\n   Plotting histograms......')
        # hierarchy of estimators: BAR > Crooks > Jarz
        if 'bar' in locals():
            show_dg = bar.dg * unit_fact
            # hierarchy of error estimates : blocks > boots > analytical
            if hasattr(bar, 'err_blocks'):
                show_err = bar.err_blocks * unit_fact
            elif hasattr(bar, 'err_boot') and not hasattr(bar, 'err_blocks'):
                show_err = bar.err_boot * unit_fact
            else:
                show_err = bar.err * unit_fact
            # plot
            plot_work_dist(fname=args.wplot, wf=res_ab, wr=res_ba, dG=show_dg,
                           dGerr=show_err, nbins=args.nbins, dpi=args.dpi,
                           units=units)
        elif 'bar' not in locals() and 'cgi' in locals():
            show_dg = cgi.dg * unit_fact
            # hierarchy of error estimates : blocks > boots
            if hasattr(cgi, 'err_blocks'):
                show_err = cgi.err_blocks * unit_fact
            elif hasattr(cgi, 'err_boot2') and not hasattr(cgi, 'err_blocks'):
                show_err = cgi.err_boot2 * unit_fact
            else:
                show_err = None
            # plot
            plot_work_dist(fname=args.wplot, wf=res_ab, wr=res_ba, dG=show_dg,
                           dGerr=show_err, nbins=args.nbins, dpi=args.dpi,
                           units=units)
        elif 'bar' not in locals() and 'cgi' not in locals() and 'jarz' in locals():
            # for the moment, show values only under specific circumstances
            if hasattr(jarz, 'dg_mean'):
                show_dg = jarz.dg_mean * unit_fact
            elif 'A' in statesProvided:
                show_dg = jarz.dg_for
            elif 'B' in statesProvided:
                show_dg = jarz.dg_rev
            else:
                show_dg = None
            show_err = None
            # plot
            plot_work_dist(fname=args.wplot, wf=res_ab, wr=res_ba, dG=show_dg,
                           dGerr=show_err, nbins=args.nbins, dpi=args.dpi,
                           units=units,statesProvided=statesProvided)

    if quiet is True:
        print('Done')
    if quiet is False:
        print('\n   ......done...........\n')

    if args.pickle and quiet is False:
        print('   NOTE: units of results in pickled files are as in the\n'
              '   provided dhdl.xvg or integ.dat files. These are typically\n'
              '   in kJ/mol when using dhdl.xvg files from Gromacs.\n')
    # execution time
    etime = time.time()
    h, m, s = _time_stats(etime-stime)
    if quiet is False:
        print("   Execution time = %02d:%02d:%02d\n" % (h, m, s))


def entry_point():
    args = parse_options()
    main(args)

if __name__ == '__main__':
    entry_point()



#!/usr/bin/env python
from argparse import ArgumentParser
import os,sys


# ==============================================================================
# CLASSES
# ==============================================================================
class ForceField:
    def __init__(self, fname=None, bForgiving=True):
        # if input is file
        if fname is not None:
            self.bForgiving = bForgiving
            self.ff = self.read_file(fname)

    def read_file(self, fname, append=False):
        # check if file exists
        if os.path.isfile(fname)==False:
            return('')

        # we store all ff info in a dict of dicts, where
        # each dict has a tuple for key and a string for value

        # if we are loading a new file/ff then we make new dicts:
        if append is False:
            ff = {}
            ff['atomtypes'] = {}
            ff['pairtypes'] = {}
            ff['bondtypes'] = {}
            ff['constrainttypes'] = {}
            ff['angletypes'] = {}
            ff['dihedraltypes'] = {}
            ff['dihedraltypesX'] = {} # dihedrals with the wildcards
        # otherwise we load the existing parameters
        elif append is True:
            ff = self.ff

        # read file lines
        lines = [x.strip().split(';')[0] for x in open(fname, 'r').readlines()
                 if x.strip().split(';')[0]]
        # remove ifdef lines: this might be an issue!
        lines = [l for l in lines if l[0] not in ('#')]

        # remove * lines
#        lines = [l for l in lines if l[0] not in ('*')]

        section = None
        dihXnum = 1 # wildcard dihedrals will be numbered
        # read each line with ff info
        for line in lines:
            # assess whether we are now parsing a different section
            if line[0] == '[':
                section = line.strip('[] ')
                continue

            # --------------------------------------------
            # cases where 1 atomtype identifies the params
            # --------------------------------------------
            if section in ('atomtypes'):
                el = line.split()
                key = (el[0])
                # check whether key is already presnet in ff
                if key not in ff[section]:
                    ff[section][key] = line
#                else:
                    # silence warning if it is about water atoms - these
                    # warnings are due to the fact we ignore ifdef statements
                    # but since the params are specific for water they are
                    # not an issue
#                    if key not in ['OW', 'HW', 'HT', 'OT', 'OWT3', 'OWT4',
#                                   'HWT3', 'HWT4', 'MWT4']:
#                        print('\n    WARNING! Atom parameters for the '
#                              ' following atoms are already present in the '
#                              'force field object'
#                              ' and will be ignored:\n    %s' % key)

            # --------------------------------------------
            # cases where 2 atomtypes identify the params
            # --------------------------------------------
            elif section in ('pairtypes', 'bondtypes', 'constrainttypes'):
                el = line.split()
                key = (el[0], el[1])
                # check whether key is already present in ff
                if key not in ff[section]:
                    ff[section][key] = line
#                else:
                    # silence some warnings: idem as above. In this case
                    # these are due to protein vsite constraints we do not
                    # care about
#                    if key not in [('MNH3', 'CT1'), ('MNH3', 'CT2'),
#                                   ('MNH3', 'CT3'), ('MNH3', 'MNH3'),
#                                   ('MCH3', 'CT1'), ('MCH3', 'CT2'),
#                                   ('MCH3', 'CT3'), ('MCH3', 'MCH3'),
#                                   ('MCH3', 'S')]:
#                        print('\n    WARNING! %s parameters for the  '
#                              'following atoms are already present in the '
#                              'force field object and will be '
#                              'ignored:\n    %s' % (section, " ".join(key)))
            # --------------------------------------------
            # cases where 3 atomtypes identify the params
            # --------------------------------------------
            elif section in ('angletypes'):
                el = line.split()
                key = (el[0], el[1], el[2])
                # check whether key is already presnet in ff
                if key not in ff[section]:
                    ff[section][key] = line
#                else:
#                    print('\n    WARNING! Angle parameters for the following '
#                          'atoms are already present in the force field object'
#                          ' and will be ignored:\n    %s' % " ".join(key))

            # --------------------------------------------
            # cases where 4 atomtypes identify the params
            # --------------------------------------------
            elif section in ('dihedraltypes'):
                el = line.split()
                dtype = int(el[4])

                # check if the dihedraltype has a wildcard
                if el[0]=='X' or el[1]=='X' or el[2]=='X' or el[3]=='X':
                    if dtype in [4, 9]:
                        # key contains: a1, a2, a3, a4, dihedral type, multiplicity
                        key = [el[0], el[1], el[2], el[3], el[4], el[7]]
                        ff['dihedraltypesX'][dihXnum] = [key,[line]]
                        dihXnum += 1
#                        print ff['dihedraltypesX'][1][0],ff['dihedraltypesX'][1][1]
#                        sys.exit(0)
                    elif dtype in [2]:
                        # key contains: a1, a2, a3, a4, dihedral type, multiplicity
                        key = [el[0], el[1], el[2], el[3], el[4]]
                        ff['dihedraltypesX'][dihXnum] = [key,[line]]
                        dihXnum += 1
                    else:
                        exit('\nBUG: dihedral type not expected/checked/'
                             'implemented yet in class ForceField. '
                             'Script needs to be extended.')
                    continue

                ####### standard dihedrals (no wildcards) ########
                # need different keys bacuse certain torsions have multiple
                # line with different multiplicities
                if dtype in [4, 9]:
                    # key contains: a1, a2, a3, a4, dihedral type, multiplicity
                    key = (el[0], el[1], el[2], el[3], el[4], el[7])
                    # check whether key is already presnet in ff
                    if key not in ff[section]:
                        ff[section][key] = line
#                    else:
#                        print('\n    WARNING! Dihedral parameters for the '
#                              'following atoms are already present in the '
#                              'force field object '
#                              'and will be ignored:\n    %s' % " ".join(key))
                elif dtype in [2]:
                    # key contains: a1, a2, a3, a4, dihedral type, multiplicity
                    key = (el[0], el[1], el[2], el[3], el[4])
                    # check whether key is already presnet in ff
                    if key not in ff[section]:
                        ff[section][key] = line
#                    else:
#                        print('\n    WARNING! Dihedral parameters for the '
#                              'following atoms are already present in the '
#                              'force field object '
#                              'and will be ignored:\n    %s' % " ".join(key))
                else:
                    exit('\nBUG: dihedral type not expected/checked/'
                         'implemented yet in class ForceField. '
                         'Script needs to be extended.')

            # --------------------------------------------
            # unkown cases
            # --------------------------------------------
            else:
                if self.bForgiving==False:
                    exit('ERROR: topology section %s not recognised/supported.' % section)
                else:
                    print('ERROR: topology section %s not recognised/supported.' % section)

        if append is False:
            return ff


# ==============================================================================
# FUNCTIONS
# ==============================================================================
def parse_options():
    parser = ArgumentParser()
    choices = ['prm', 'ff']
    parser.add_argument('-itp', dest='itp',
                        help='Gromacs topology, containing CGenFF parameters '
                        ' for a ligand obtained from the CGenFF program '
                        '(paramchem) and converted with the '
                        '"cgenff_charmm2gmx.py" script.',
                        required=True)
    parser.add_argument('-ff', dest='ff',
                        help='Path to the Charmm36 force '
                        'field, which contains the CGenFF parameters.',
                        required=True)
    parser.add_argument('-prm', dest='prm',
                        help='File with additional parameters that were '
                        'assigned by analogy by the CGenFF program '
                        '(paramchem). (This prm is in gromacs format analogous to .itp)',
                        required=False)
    parser.add_argument('-o', dest='outfname',
                        help='Output itp file which contains all of the force '
                        'field parameters explicitly (excluding the atomtypes'
                        '). Default is "cgenff.itp"',
                        default='cgenff.itp')
    parser.add_argument('--priority', dest='priority',
                        help='This determines the behavior when dealing with '
                        'duplicate parameters. Options are "prm" or "ff" '
                        '(default is "ff"). With "ff" the parameters already in'
                        'the force field take priority and are used; with '
                        '"prm" the parameters present in the prm file are '
                        'used instead in case of duplicate entries. The '
                        'latter option is useful if you have optimized some '
                        'terms specifically for the molecule at hand, e.g. '
                        'with the Force Field Toolkit.',
                        default='ff', choices=choices)
    parser.add_argument('--bForgiving', dest='bForgiving',
                        help='Should the script exit when some ff section is not recognized.',
                        default=False)
    args = parser.parse_args()
    return args


def list2file(lst, outfname):
    with open(outfname, 'w') as f:
        for l in lst:
            f.write('{line}\n'.format(line=l))


def get_atoms_map(itp):
    # read the atoms from the itp file
    lines = [x.strip().split(';')[0] for x in open(itp, 'r').readlines()
             if x.strip().split(';')[0]]

    # map atom ids to their atomtypes
    section = ''
    id2type_map = {}
    for line in lines:
        # assess whether we are now parsing a different section
        if line[0] == '[':
            section = line.strip('[] ')
            continue
        if section == 'atoms':
            el = line.split()
            a_id = el[0]
            a_type = el[1]
            id2type_map[a_id] = a_type

    return id2type_map

def check_wildcard_dihedral( key, at ):
    found = True
    for k,a in zip(key,at):
        if k=='X':
            continue
        elif k==a:
            continue
        else:
            found=False
            return(found)
    return(found)

def fill_params(itp, a_map, ff):
    # read the atoms from the itp file
    lines = [x.strip().split(';')[0] for x in open(args.itp, 'r').readlines()
             if x.strip().split(';')[0]]

    out = []

    section = ''
    for line in lines:
        if line.startswith('#'):
            continue

        # assess whether we are now parsing a different section
        if line[0] == '[':
            section = line.strip('[] ')
            out.append(line)
            continue

        # -----
        # Bonds
        # -----
        if section == 'bonds':
            el = line.split()
            a1 = a_map[el[0]]
            a2 = a_map[el[1]]
            key = (a1, a2)

            # order does not matter: check both
            try:
                params = ff['bondtypes'][key]
            except:
                key = key[::-1]
                params = ff['bondtypes'][key]

            # remove the atomtypes from the ff string
            paramsplit = params.split()
            params = ""
            for p in paramsplit:
                if p not in key:
                    params += " "+str(p)
#            for k in key:
                #params = params.replace(k, '')
		#print k," --- VG ---",params

            # create new topology line
            newline = '{a1:<5}{a2:<5}{params}'.format(a1=el[0],
                                                      a2=el[1],
                                                      params=params)
            out.append(newline)

        # ------
        # Angles
        # ------
        elif section == 'angles':
            el = line.split()
            a1 = a_map[el[0]]
            a2 = a_map[el[1]]
            a3 = a_map[el[2]]
            key = (a1, a2, a3)

            # order does not matter: check both
            try:
                params = ff['angletypes'][key]
            except:
                key = key[::-1]
                params = ff['angletypes'][key]

            # remove the atomtypes from the ff string
            paramsplit = params.split()
            params = ""
            for p in paramsplit:
                if p not in key:
                    params += " "+str(p)
#            for k in key:
#                params = params.replace(k, '')

            # create new topology line
            newline = '{a1:<5}{a2:<5}{a3:<5}{params}'.format(a1=el[0],
                                                             a2=el[1],
                                                             a3=el[2],
                                                             params=params)
            out.append(newline)

        # ---------
        # Dihedrals
        # ---------
        elif section == 'dihedrals':
            el = line.split()
            a1 = a_map[el[0]]
            a2 = a_map[el[1]]
            a3 = a_map[el[2]]
            a4 = a_map[el[3]]
            dtype = el[4]

            # get largest multiplicity in ff
            max_pn = max([int(x[-1]) for x in ff['dihedraltypes'].keys()])

            # -----------------------------
            # dihedrals with multiplicity
            # -----------------------------
            if int(dtype) in [4, 9]:
                found = False
                # check various multiplicities
                for pn in range(1, max_pn+1):
                    # check both orders
                    try:
                        key = (a1, a2, a3, a4, dtype, str(pn))
                        params = ff['dihedraltypes'][key]
                        found = True
                    except:
                        try:
                            key = (a4, a3, a2, a1, dtype, str(pn))
                            params = ff['dihedraltypes'][key]
                            found = True
                        except:
                            # if missing it's because there is no param
                            # with this multiplicity. This can however
                            # be dangerous, so we use "found" to make sure
                            # the term is not skipped (i.e. at least one
                            # torsion needs to be found)
                            continue

                    if found==True:
                        # remove the atomtypes from the ff string
                        paramsplit = params.split()
                        params = ""
                        for p in paramsplit:
                            if p not in [a1,a2,a3,a4]:
                                params += " "+str(p)
                        # create new topology line
                        newline = ('{a1:<5}{a2:<5}{a3:<5}{a4:<5}'
                                   '{params}'.format(a1=el[0], a2=el[1],
                                                     a3=el[2], a4=el[3],
                                                     params=params))
                        out.append(newline)


		############################################################
		# if not yet found a dihedral, search through the wildcards #
                # the search depends on the dihedral number #
                # the first match (or matches of different multiplicities will be used #
                if found==False:
                    dihXparams = []
                    prev = 0
                    for d in ff['dihedraltypesX']:
                        # first check that the dihedral is of the correct type
                        if int(ff['dihedraltypesX'][d][0][4]) not in [4,9]:
                            continue
                        # in case another dihedral matching the wildcard is found,
                        # but it is not directly after the previous dihedral
                        if found==True and d!=prev+1:
                            break
                        if check_wildcard_dihedral( ff['dihedraltypesX'][d][0], [a1,a2,a3,a4] )==True:
                            params = ff['dihedraltypesX'][d][1]
                            dihXparams.append(params)
                            found = True
                            prev = d
                        elif check_wildcard_dihedral( ff['dihedraltypesX'][d][0], [a4,a3,a2,a1] )==True:
                            params = ff['dihedraltypesX'][d][1]
                            dihXparams.append(params)
                            found = True
                            prev = d

                    # remove the atomtypes from the ff string
                    for params in dihXparams:
                        paramsplit = params[0].split()
                        params = ""
                        for p in paramsplit:
                            if p not in [a1,a2,a3,a4]:
                                if p!='X':
                                    params += " "+str(p)
                        # create new topology line
                        newline = ('{a1:<5}{a2:<5}{a3:<5}{a4:<5}'
                                   '{params}'.format(a1=el[0], a2=el[1],
                                                     a3=el[2], a4=el[3],
                                                     params=params))
                        out.append(newline)


                # check whether at least one torsional term was found
                if found==False:
                    print('\n    ERROR! Could not find dihedral parameters:\n %s' % " ".join(key))
#                              ' following atoms are already present in the '
#                              'force field object'
#                              ' and will be ignored:\n    %s' % key)
                    sys.exit(0)
                #assert found is True

            # -----------------------------
            # dihedrals without multiplicity
            # -----------------------------
            elif int(dtype) in [2]:
                found = False
                el = line.split()
                a1 = a_map[el[0]]
                a2 = a_map[el[1]]
                a3 = a_map[el[2]]
                a4 = a_map[el[3]]
                dtype = el[4]

                # order does not matter: check both
                try:
                    key = (a1, a2, a3, a4, dtype)
                    params = ff['dihedraltypes'][key]
                    found = True
                except:
                    try:
                        key = (a4, a3, a2, a1, dtype)
                        params = ff['dihedraltypes'][key]
                        found = True
                    except:
                        found = False

                if found==True:
                    # remove the atomtypes from the ff string
                    paramsplit = params.split()
                    params = ""
                    for p in paramsplit:
                        if p not in [a1,a2,a3,a4]:
                            params += " "+str(p)
                    # create new topology line
                    newline = ('{a1:<5}{a2:<5}{a3:<5}{a4:<5}'
                               '{params}'.format(a1=el[0], a2=el[1],
                                                 a3=el[2], a4=el[3],
                                                 params=params))
                    out.append(newline)

                ############################################################
                # if not yet found a dihedral, search through the wildcards #
                # the search depends on the dihedral number #
                # the first match will be the correct dihedral #
                if found==False:
                    dihXparams = ''
                    for d in ff['dihedraltypesX']:
                        # first check that the dihedral is of the correct type
                        if int(ff['dihedraltypesX'][d][0][4]) not in [2]:
                            continue
                        if check_wildcard_dihedral( ff['dihedraltypesX'][d][0], [a1,a2,a3,a4] )==True:
                            dihXparams = ff['dihedraltypesX'][d][1]
                            found = True
                            break
                        elif check_wildcard_dihedral( ff['dihedraltypesX'][d][0], [a4,a3,a2,a1] )==True:
                            dihXparams = ff['dihedraltypesX'][d][1]
                            found = True
                            break

                    # remove the atomtypes from the ff string
                    paramsplit = dihXparams[0].split()
#                    print paramsplit
#                    sys.exit(0)
                    params = ""
                    for p in paramsplit:
                        if p not in [a1,a2,a3,a4]:
                            if p!='X':
                                params += " "+str(p)
                    # create new topology line
                    newline = ('{a1:<5}{a2:<5}{a3:<5}{a4:<5}'
                               '{params}'.format(a1=el[0], a2=el[1],
                                                 a3=el[2], a4=el[3],
                                                 params=params))
                    out.append(newline)

            # ----------------------
            # catch unkown dihedrals
            # ----------------------
            else:
                exit('\nBUG: dihedral type %s not expected/checked/'
                     'implemented yet in func "fill_params".'
                     ' Script needs to be extended.' % dtype)

        # -----------------------------------------------------------
        # Just copy the other lines, e.g. the pairs, moleculetype etc.
        # -----------------------------------------------------------
        else:
            out.append(line)

    return out


# ==============================================================================
# MAIN
# ==============================================================================
def main(args):
    print('\n Reading force field information...')

    if args.priority == 'ff':
        if 'itp' in args.ff:
            ffparams = ForceField(fname='{path}'.format(path=args.ff), bForgiving=args.bForgiving)
            ffparams.read_file(fname='{path}'.format(path=args.ff),append=True)
        else:
            ffparams = ForceField(fname='{path}/ffnonbonded.itp'.format(path=args.ff), bForgiving=args.bForgiving)
            ffparams.read_file(fname='{path}/ffbonded.itp'.format(path=args.ff),append=True)
        ffparams.read_file(fname='{prm}'.format(prm=args.prm),append=True)
    elif args.priority == 'prm':
        ffparams = ForceField(fname='{prm}'.format(prm=args.prm), bForgiving=args.bForgiving)
        ffparams.read_file(fname='{path}/ffnonbonded.itp'.format(path=args.ff),
                              append=True)
        ffparams.read_file(fname='{path}/ffbonded.itp'.format(path=args.ff),
                           append=True)

    print('\n Filling in the topology file...')
    # map atom ids to their atomtypes
    id2type_map = get_atoms_map(itp=args.itp)
    print(id2type_map)
    # put all ff parameters explicitly in the itp file
    newitp = fill_params(itp=args.itp, a_map=id2type_map, ff=ffparams.ff)
    # save the new itp
    list2file(lst=newitp, outfname=args.outfname)
    print(' New topology file saved as "{fname}"\n'.
          format(fname=args.outfname))

# call main
if __name__ == "__main__":
    args = parse_options()
    main(args)



#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""This file contains the Rotation class.
A Rotation instance is built with two vectors.

Examples
--------

    >>> v1 = [1,2,3]
    >>> v2 = [2,3,4]
    >>> r = Rotation(v1,v2)  # create rotation object around v2-v1
    >>> v3 = [4,5,6]
    >>> v3 = r.apply(v3)     # rotate v3 around v2-v1

"""

from numpy import array, linalg, arccos, inner
from . import _pmx as _p


class Rotation:

    def __init__(self, v1, v2):
        """ creates a rotation object
        around the vector v2-v1"""

        self.v1 = array(v1)
        self.v2 = [v2[0], v2[1], v2[2]]
        tmp = array(v2)
        self.vec = tmp-self.v1
        self.m1 = None
        self.m2 = None
        x = 1./linalg.norm(self.vec)
        self.norm_vec = self.vec*x
        self.__rm1()
        self.__rm2()

    def __rm1(self):

        a = self.norm_vec
        self.m1 = [
            [a[0]*a[0], a[0]*a[1], a[0]*a[2]],
            [a[1]*a[0], a[1]*a[1], a[1]*a[2]],
            [a[2]*a[0], a[2]*a[1], a[2]*a[2]]
            ]

    def __rm2(self):

        a = self.norm_vec
        self.m2 = [
            [0.0, -a[2], a[1]],
            [a[2], 0.0, -a[0]],
            [-a[1], a[0], 0.0]
            ]

    def apply(self, v, phi):
        return _p.apply_rotation(self, [v[0], v[1], v[2]], phi)


def vec_ang(v1, v2):
    x1 = linalg.norm(v1)
    x2 = linalg.norm(v2)
    return arccos(inner(v1, v2)/(x1*x2))


def bb_super(mol1, mol2, use_orig_mc_coords=True):
    """ superpose mol2 on mol1"""

    N1, CA1, C1 = mol1.fetchm(['N', 'CA', 'C'])
    N2, CA2, C2 = mol2.fetchm(['N', 'CA', 'C'])

    if ((mol1.resname == 'GLY') or (mol2.resname == 'GLY') or
       (mol2.resname[:2] == 'G2')):
        fit_atoms([N1, CA1, C1], [N2, CA2, C2], mol2.atoms)
    else:
        N1, CA1, C1, CB1 = mol1.fetchm(['N', 'CA', 'C', 'CB'])
        N2, CA2, C2, CB2 = mol2.fetchm(['N', 'CA', 'C', 'CB'])
        fit_atoms([N1, CA1, C1, CB1], [N2, CA2, C2, CB2], mol2.atoms)

    if use_orig_mc_coords:
        atom_set = ['N', 'CA', 'C', 'H', 'O', 'HA', 'HN']
        gly_atom_set = ['N', 'CA', 'C', 'H', 'O', 'HA1', 'HN']
        pro_atom_set = ['N', 'CA', 'C', 'O', 'HA']
        pro_atom_set_forGly = ['N', 'CA', 'C', 'O', 'HA1']

        # for proline it is different
        if mol1.resname=='PRO' or mol2.resname == 'PRO' or mol2.resname[:2] == 'P2':
            atoms1 = mol1.fetchm(pro_atom_set)
            atoms2 = mol2.fetchm(pro_atom_set)
            if mol1.resname=='GLY':
                atoms1 = mol1.fetchm(pro_atom_set_forGly)
            elif mol2.resname=='GLY' or mol2.resname[:2] == 'G2':
                atoms2 = mol2.fetchm(pro_atom_set_forGly)
        else:
            if mol1.resname == 'GLY':
                atoms1 = mol1.fetchm(gly_atom_set)
            else:
                atoms1 = mol1.fetchm(atom_set)

            if mol2.resname == 'GLY' or mol2.resname[:2] == 'G2':
                atoms2 = mol2.fetchm(gly_atom_set)
            else:
                atoms2 = mol2.fetchm(atom_set)

        assert len(atoms1) == len(atoms2), "%s -> %s" % ('-'.join(map(lambda a: a.name, atoms1)), '-'.join(map(lambda a: a.name, atoms2)))
        for atom1, atom2 in zip(atoms1, atoms2):
            atom2.x = atom1.x


def nuc_super(mol1, mol2, name1=None, name2=None):
    """ superpose mol2 on mol1"""

    if name1 is None:
        name1 = mol1.resname[:2]
    if name2 is None:
        name2 = mol2.resname[:2]

    if name1 in ['DT', 'DC', 'RC', 'RU']:
        fit1_atoms = ['C1\'', 'C6', 'N1', 'C2', 'C5', 'N3']
    else:
        fit1_atoms = ['C1\'', 'C8', 'N9', 'C4', 'N7', 'C5']

    if name2 in ['DT', 'DC', 'RC', 'RU']:
        fit2_atoms = ['C1\'', 'C6', 'N1', 'C2', 'C5', 'N3']
    else:
        fit2_atoms = ['C1\'', 'C8', 'N9', 'C4', 'N7', 'C5']

    atoms1 = mol1.fetchm(fit1_atoms)
    atoms2 = mol2.fetchm(fit2_atoms)

    fit_atoms(atoms1, atoms2, mol2.atoms)


def planarity(atom_list):

    coords = list(map(lambda a: a.x, atom_list))
    plan = _p.planarity(coords)
    return plan


def apply_fit_R(atoms, R):

    for atom in atoms:
        x_old = list(map(lambda x: x, atom.x))
        for r in range(3):
            atom.x[r] = 0
            for c in range(3):
                atom.x[r] += R[r][c]*x_old[c]

def center_vector( v ):
    vout = _p.center_vec( v )
    return( vout )

def calc_fit_R( cs1, cs2, m ):
    R = _p.calc_fit_R(cs1, cs2, m)
    return(R)


def fit(model1, model2, atom_names=[]):
    if atom_names:
        subset1 = model1.fetch_atoms(atom_names)
        subset2 = model2.fetch_atoms(atom_names)
        cs1 = list(map(lambda a: a.x, subset1))
        cs2 = list(map(lambda a: a.x, subset2))
    else:
        cs1 = model1.coords()
        cs2 = model2.coords()

    assert(len(cs1) == len(cs2))
    m = list(map(lambda x: 1., cs1))  # dummy array
    v = _p.center_vec(cs1)
    v2 = _p.center_vec(cs2)
    R = _p.calc_fit_R(cs1, cs2, m)
    model2.translate([-v2[0], -v2[1], -v2[2]])
    apply_fit_R(model2.atoms, R)
    model2.translate(v)


def fit_by_ndx(ref, model, ndx1, ndx2):
    crd1 = list(map(lambda i: ref.atoms[i-1].x, ndx1))
    crd2 = list(map(lambda i: model.atoms[i-1].x, ndx2))

    assert(len(crd1) == len(crd2))
    m = list(map(lambda x: 1., crd1))  # dummy array
    v = _p.center_vec(crd1)
    v2 = _p.center_vec(crd2)
    R = _p.calc_fit_R(crd1, crd2, m)
    model.translate([-v2[0], -v2[1], -v2[2]])
    apply_fit_R(model.atoms, R)
    model.translate(v)


def translate_by_ndx(struct, ndx):
    crd = list(map(lambda i: struct.atoms[i-1].x, ndx))
    m = list(map(lambda x: 1., crd))  # FIXME: variable m is not used
    v = _p.center_vec(crd)
    struct.translate([-v[0], -v[1], -v[2]])
    return(v)


def fit_atoms(fit_atoms1, fit_atoms2, rot_atoms2):

    cs1 = list(map(lambda a: a.x, fit_atoms1))
    cs2 = list(map(lambda a: a.x, fit_atoms2))
    assert len(cs1) == len(cs2)
    m = list(map(lambda x: 1., cs1))  # dummy array
    v = _p.center_vec(cs1)
    v2 = _p.center_vec(cs2)
    R = _p.calc_fit_R(cs1, cs2, m)
    for atom in rot_atoms2:
        atom.x[0] -= v2[0]
        atom.x[1] -= v2[1]
        atom.x[2] -= v2[2]
    apply_fit_R(rot_atoms2, R)
    for atom in rot_atoms2:
        atom.x[0] += v[0]
        atom.x[1] += v[1]
        atom.x[2] += v[2]



#!/usr/bin/env python

"""This module contains classes for calling rosetta scripts:
Flex_DDG, ...
"""

import os
import sys
from copy import deepcopy
import numpy as np
from . import library
from .atom import Atom
from .model import Model
from .resselection import InteractiveSelection
from pmx import resselection
from pmx.parser import parseList, kickOutComments
from .utils import create_folder,remove_netmount,importError
from pmx import jobscripts
import sqlite3, shutil,tempfile,re,datetime,collections
from pprint import pprint
import pandas as pd
from pmx.rosetta import muts_read_and_format

try:
    import sqlite3
except:
    raise(importError('sqlite3'))
    

class FlexDDG_analyze:
    '''Analysis of the results from the Flex_ddg protocol 
       for protein-protein binding affinity changes
       upon amino mutation [Barlow et al, JPCB, 2018].


    Parameters
    ----------
    mutFile : str, optional
        file name with mutations (chain resID mut)
    mutList : list, optional
        list of mutations
    path : str, optional
        path where the mutations are located.
    outfile : str, optional
        output file with all results summarized.
    overwrite: bool, optional
        if already present, should the results file
        be overwritten. Default: False
    nstruct : int,
        number of runs per mutation.
    
    '''
#MIT License

#Copyright (c) 2018 Kyle Barlow

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

    def __init__( self, mutFile=None, mutList=[], path='.', outfile='', overwrite=False, nstruct=35 ):#, backrub_trajectory_stride=35000 ):

        # get parameters
        self.basepath = path
        self.outfile = outfile
        self.bOverwrite = overwrite
        self.nstruct = nstruct
        self.trajectory_stride = 35#backrub_trajectory_stride
        # other default rosetta parameters
        self.rosetta_output_file_name = 'rosetta.out'
        self.output_database_name = 'ddG.db3'
        self.script_output_folder = 'analysis_output'

        # get folders
        self.folders = []
        if (mutFile is None) and len(mutList)==0:
            self._mutations_from_folders( )
        elif mutFile is not None:
            self.sele = muts_read_and_format( mutFile, 'sss' )
            for sel in self.sele:
                folder = 'mut_ch{0}_{1}{2}'.format(sel[0][0],sel[0][1],sel[0][2])
                for i in range(1,len(sel)): # several mutations simultaneously
                    s = sel[i]
                    folder = folder+'_ch{0}_{1}{2}'.format(s[0],s[1],s[2])
                self.folders.append( folder )
        self.folderNum = len(self.folders)

        # output file
        self.outfile = outfile
        self.alreadyProcessedFolders = []
        self._process_outfile()

        # gam weighting params
        self.zemu_gam_params = {
            'fa_sol' :      (6.940, -6.722),
            'hbond_sc' :    (1.902, -1.999),
            'hbond_bb_sc' : (0.063,  0.452),
            'fa_rep' :      (1.659, -0.836),
            'fa_elec' :     (0.697, -0.122),
            'hbond_lr_bb' : (2.738, -1.179),
            'fa_atr' :      (2.313, -1.649),
        }

        # analyze every mutation
        counter = 1
        for folder in self.folders:
            if folder in self.alreadyProcessedFolders:
                print('Already processed: {0}/{1} {2} '.format(counter,self.folderNum,folder))
                counter+=1
                continue
            print('Analyzing: {0}/{1} {2} '.format(counter,self.folderNum,folder))
            # get the file paths
            finished_jobs = self._find_finished_jobs( self.basepath+'/'+folder )
            if len(finished_jobs.keys())==0:
                print('****Failed (no finished jobs found): {0}/{1} {2}****'.format(counter,self.folderNum,folder))
                counter+=1
                continue

            # get the values
            ddg_scores_dfs = []
            struct_scores_dfs = []
            for key in finished_jobs.keys():
                finished_job = finished_jobs[key]
                inner_scores_list = []
                inner_scores = self._process_finished_struct( finished_job, os.path.basename(finished_job) )
                inner_scores_list.append( inner_scores )
                scores = pd.concat( inner_scores_list )
                ddg_scores, struct_scores = self._calc_ddg( scores )
                struct_scores_dfs.append( struct_scores )
                ddg_scores_dfs.append( ddg_scores )
                ddg_scores_dfs.append( self._apply_zemu_gam(ddg_scores) )
                ddg_scores_dfs.extend( self._calc_dgs( scores ) )

#            pd.concat( struct_scores_dfs ).to_csv( 'tmp.csv' ) 
#            df = pd.concat( ddg_scores_dfs, sort=True )
#            df.to_csv( os.path.join('tmp2.csv') )

            # extract ddG and calculate mean, sderr
            df = pd.concat( ddg_scores_dfs,sort=True )
            ddgvalsGAM = df.loc[ (df['scored_state']=='ddG') & (df['score_function_name'].str.contains('-gam')) ]['total_score']
            ddgvals = df.loc[ (df['scored_state']=='ddG') & (df['score_function_name'].str.contains('-gam')==False) ]['total_score']
            # w/o -gam
            meanval = np.mean(ddgvals)
            sderrval = np.std(ddgvals,ddof=1)/np.sqrt(float(np.shape(ddgvals)[0]))
            # w/ -gam
            meanvalGAM = np.mean(ddgvalsGAM)
            sderrvalGAM = np.std(ddgvalsGAM,ddof=1)/np.sqrt(float(np.shape(ddgvalsGAM)[0]))
            # output
            outfp = open(self.outfile,'a')
            outfp.write( '%10s | %3.2f %3.2f | %3.2f %3.2f\n' % (folder,np.round(meanval,2),np.round(sderrval,2),np.round(meanvalGAM,2),np.round(sderrvalGAM,2)) )
            outfp.close()

            counter+=1


    def _process_outfile( self ):
        if self.outfile=='':
            self.outfile = self.basepath+'/results.dat'
        if self.bOverwrite==True:
            fp = open(self.outfile,'w')
            fp.close()
        # if file already exists, read its contents and append to them
        if os.path.isfile(self.outfile):
            fp = open(self.outfile,'r')
            for l in fp.readlines():
                l = l.rstrip()
                foo = l.split()
                if foo[0].startswith('mut_'):
                    self.alreadyProcessedFolders.append(foo[0])
            fp.close()

    def _gam_function(self, x, score_term = None ):
        return -1.0 * np.exp( self.zemu_gam_params[score_term][0] ) + 2.0 * np.exp( self.zemu_gam_params[score_term][0] ) / ( 1.0 + np.exp( -1.0 * x * np.exp( self.zemu_gam_params[score_term][1] ) ) )

    def _apply_zemu_gam(self, scores):
        new_columns = list(scores.columns)
        new_columns.remove('total_score')
        scores = scores.copy()[ new_columns ]
        for score_term in self.zemu_gam_params:
            assert( score_term in scores.columns )
            scores[score_term] = scores[score_term].apply( self._gam_function, score_term = score_term )
        scores[ 'total_score' ] = scores[ list(self.zemu_gam_params.keys()) ].sum( axis = 1 )
        scores[ 'score_function_name' ] = scores[ 'score_function_name' ] + '-gam'
#        self.gam_function_name = scores[ 'score_function_name' ] + '-gam'
#        self.score_function_name = scores['score_function_name']
        return(scores)

    def _mutations_from_folders( self ):
        folderContents = os.listdir(self.basepath)
        
        for folder in folderContents:
            if folder.startswith('mut_'):
                self.folders.append(folder)

    def _rosetta_output_succeeded( self,potential_struct_dir ):
        path_to_rosetta_output = os.path.join( potential_struct_dir, self.rosetta_output_file_name )
        if not os.path.isfile(path_to_rosetta_output):
            return False

        db3_file = os.path.join( potential_struct_dir, self.output_database_name )
        if not os.path.isfile( db3_file ):
            return False

        success_line_found = False
        no_more_batches_line_found = False
        with open( path_to_rosetta_output, 'r' ) as f:
            for line in f:
                if line.startswith( 'protocols.jd2.JobDistributor' ) \
                   and ( ('reported success in' in line) \
                   or ('did you forget to pass -overwrite' in line) ):
                    success_line_found = True
                if line.startswith( 'protocols.jd2.JobDistributor' ) and 'no more batches to process' in line:
                    no_more_batches_line_found = True

        return(no_more_batches_line_found and success_line_found)

    def _find_finished_jobs( self,folder ):
        return_dict = {}
        for i in range(1, self.nstruct+1):
            job_dir = folder+'/'+str(i)
            if i<10:
                job_dir = folder+'/0'+str(i)
            job_dir = os.path.abspath(job_dir)
            if os.path.isdir(job_dir):
                if self._rosetta_output_succeeded( job_dir ):
                    return_dict[i] = job_dir
        return(return_dict)

    def _process_finished_struct( self, output_path, case_name ):
        db3_file = os.path.join( output_path, self.output_database_name )
        assert( os.path.isfile( db3_file ) )
        struct_number = int( os.path.basename(output_path) )
        scores_df = self._get_scores_from_db3_file( db3_file, struct_number, case_name )
        return(scores_df)

    def _get_scores_from_db3_file(self, db3_file, struct_number, case_name):
        conn = sqlite3.connect(db3_file)
        conn.row_factory = sqlite3.Row
        c = conn.cursor()

        num_batches = c.execute('SELECT max(batch_id) from batches').fetchone()[0]
        scores = pd.read_sql_query('''
        SELECT batches.name, structure_scores.struct_id, score_types.score_type_name, structure_scores.score_value, score_function_method_options.score_function_name from structure_scores
        INNER JOIN batches ON batches.batch_id=structure_scores.batch_id
        INNER JOIN score_function_method_options ON score_function_method_options.batch_id=batches.batch_id
        INNER JOIN score_types ON score_types.batch_id=structure_scores.batch_id AND score_types.score_type_id=structure_scores.score_type_id
        ''', conn)

        def renumber_struct_id( struct_id ):
            return self.trajectory_stride * ( 1 + (int(struct_id-1) // num_batches) )

        scores['struct_id'] = scores['struct_id'].apply( renumber_struct_id )
        scores['name'] = scores['name'].apply( lambda x: x[:-9] if x.endswith('_dbreport') else x )
        scores = scores.pivot_table( index = ['name', 'struct_id', 'score_function_name'], columns = 'score_type_name', values = 'score_value' ).reset_index()
        scores.rename( columns = {
            'name' : 'state',
            'struct_id' : 'backrub_steps',
        }, inplace=True)
        scores['struct_num'] = struct_number
        scores['case_name'] = case_name
        conn.close()
        return(scores)

    def _calc_ddg( self, scores ):
        total_structs = np.max( scores['struct_num'] )

        nstructs_to_analyze = set([total_structs])
        for x in range(10, total_structs):
            if x % 10 == 0:
                nstructs_to_analyze.add(x)
        nstructs_to_analyze = sorted(nstructs_to_analyze)

        all_ddg_scores = []
        for nstructs in nstructs_to_analyze:
            ddg_scores = scores.loc[ ((scores['state'] == 'unbound_mut') | (scores['state'] == 'bound_wt')) & (scores['struct_num'] <= nstructs) ].copy()
            for column in ddg_scores.columns:
                if column not in ['state', 'case_name', 'backrub_steps', 'struct_num', 'score_function_name']:
                    ddg_scores.loc[:,column] *= -1.0
            ddg_scores = ddg_scores.append( scores.loc[ ((scores['state'] == 'unbound_wt') | (scores['state'] == 'bound_mut')) & (scores['struct_num'] <= nstructs) ].copy() )
            ddg_scores = ddg_scores.groupby( ['case_name', 'backrub_steps', 'struct_num', 'score_function_name'] ).sum().reset_index()

            if nstructs == total_structs:
                struct_scores = ddg_scores.copy()

            ddg_scores = ddg_scores.groupby( ['case_name', 'backrub_steps', 'score_function_name'] ).mean().round(decimals=5).reset_index()
            new_columns = list(ddg_scores.columns.values)
            new_columns.remove( 'struct_num' )
            ddg_scores = ddg_scores[new_columns]
            ddg_scores[ 'scored_state' ] = 'ddG'
            ddg_scores[ 'nstruct' ] = nstructs
            all_ddg_scores.append(ddg_scores)
        return(pd.concat(all_ddg_scores), struct_scores)

    def _calc_dgs( self,scores ):
        l = []

        total_structs = np.max( scores['struct_num'] )

        nstructs_to_analyze = set([total_structs])
        for x in range(10, total_structs):
            if x % 10 == 0:
                nstructs_to_analyze.add(x)
        nstructs_to_analyze = sorted(nstructs_to_analyze)

        for state in ['mut', 'wt']:
            for nstructs in nstructs_to_analyze:
                dg_scores = scores.loc[ (scores['state'].str.endswith(state)) & (scores['state'].str.startswith('unbound')) & (scores['struct_num'] <= nstructs) ].copy()
                for column in dg_scores.columns:
                    if column not in ['state', 'case_name', 'backrub_steps', 'struct_num', 'score_function_name']:
                        dg_scores.loc[:,column] *= -1.0
                dg_scores = dg_scores.append( scores.loc[ (scores['state'].str.endswith(state)) & (scores['state'].str.startswith('bound')) & (scores['struct_num'] <= nstructs) ].copy() )
                dg_scores = dg_scores.groupby( ['case_name', 'backrub_steps', 'struct_num', 'score_function_name'] ).sum().reset_index()
                dg_scores = dg_scores.groupby( ['case_name', 'backrub_steps', 'score_function_name'] ).mean().round(decimals=5).reset_index()
                new_columns = list(dg_scores.columns.values)
                new_columns.remove( 'struct_num' )
                dg_scores = dg_scores[new_columns]
                dg_scores[ 'scored_state' ] = state + '_dG'
                dg_scores[ 'nstruct' ] = nstructs
                l.append( dg_scores )
        return(l)
